from django.contrib import admin, messages
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from wbcore.contrib.authentication.admin import UserAdmin

from wbcrm.admin import ActivityAdmin
from wbcrm.models import Activity

from .shortcuts import get_backend

User = get_user_model()
admin.site.unregister(User)
admin.site.unregister(Activity)


@admin.register(Activity)
class ActivitySyncAdmin(ActivityAdmin):
    def delete_queryset(self, request, queryset):
        """Given a queryset, delete it from the database."""
        for obj in queryset.filter(is_active=True):
            obj.delete()
        super().delete_queryset(request, queryset)


@admin.register(User)
class UserSyncAdmin(UserAdmin):
    def set_web_hook(self, request, queryset):
        try:
            if controller := get_backend():
                for user in queryset:
                    controller.backend.set_web_hook(user)
                self.message_user(
                    request,
                    _("Operation completed, we have set the webhook for {} users.").format(queryset.count()),
                )
            else:
                raise ValueError("No backend set in preferences")
        except Exception as e:
            self.message_user(request, _("Operation Failed, {}").format(e), messages.WARNING)

    def stop_web_hook(self, request, queryset):
        try:
            if controller := get_backend():
                for user in queryset:
                    controller.backend.stop_web_hook(user)
                self.message_user(
                    request,
                    _("Operation completed, we have stopped the webhook for {} users.").format(queryset.count()),
                )
            else:
                raise ValueError("No backend set in preferences")
        except Exception as e:
            self.message_user(request, _("Operation Failed, {}").format(e), messages.WARNING)

    def check_web_hook(self, request, queryset):
        try:
            if controller := get_backend():
                for user in queryset:
                    controller.backend.check_web_hook(user)
                self.message_user(
                    request,
                    _("Operation completed, we checked the webhook for {} users.").format(queryset.count()),
                )
            else:
                raise ValueError("No backend set in preferences")
        except Exception as e:
            self.message_user(request, _("Operation Failed, {}").format(e), messages.WARNING)

    actions = UserAdmin.actions + (
        set_web_hook,
        stop_web_hook,
        check_web_hook,
    )



import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory

from wbcrm.report.activity_report import create_report, create_report_and_send


@pytest.mark.django_db
class TestSpecificReport:
    def test_create_report(self, person_factory, activity_factory):
        person = person_factory()
        activity = activity_factory(participants=(person,))
        report = create_report(person.id, activity.start, activity.end)
        assert report
        assert report.getvalue()

    def test_create_report_and_send(self, person_factory):
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        person = person_factory()
        create_report_and_send(request.user.profile.id, person.id)



# Generated by Django 4.2.11 on 2024-03-26 10:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0008_alter_activity_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="is_public",
            field=models.BooleanField(
                default=True, help_text="If True, all internal users can access this account", verbose_name="Public"
            ),
        ),
    ]



from django.contrib.messages import warning
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework import permissions
from rest_framework.decorators import action
from rest_framework.response import Response

from wbcrm.models.recurrence import Recurrence


class RecurrenceModelViewSetMixin:
    @action(detail=True, methods=["DELETE"], permission_classes=[permissions.IsAuthenticated])
    def delete_next_occurrences(self, request, pk=None):
        instance = get_object_or_404(self.queryset.model, pk=pk)
        instance.forward_deletion()
        return Response({})

    def add_messages(self, request, instance=None, **kwargs):
        if instance:
            # If true, this activity is recurring.
            if instance.is_recurrent:
                warning(
                    request,
                    _("This is a recurring occurrence with period {repeat_choice}").format(
                        repeat_choice=Recurrence.ReoccuranceChoice(instance.repeat_choice).label
                    ),
                )



from datetime import datetime
from io import BytesIO

import xlsxwriter
from celery import shared_task
from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.db.models import Q
from django.template.loader import get_template
from django.utils import timezone
from django.utils.html import strip_tags
from django.utils.translation import gettext as _
from wbcore.contrib.directory.models import Person
from wbcrm.models import Activity, ActivityType
from xlsxwriter.utility import xl_rowcol_to_cell


def set_cell(worksheet, row, col, value, width_cols, base_format):
    cell = xl_rowcol_to_cell(row, col)
    if value:
        worksheet.write_string(cell, value, base_format)
        if len(value) > width_cols[col]:
            width_cols[col] = len(value)


@shared_task
def create_report_and_send(profile_id, employee_id, start_date=None, end_date=None):
    employee = Person.all_objects.get(id=employee_id)
    profile = Person.all_objects.get(id=profile_id)
    if end_date is None:
        end_date = timezone.now()
    if start_date is None:
        start_date = datetime(year=end_date.year, month=1, day=1, hour=0, minute=0, second=0)

    report = create_report(employee_id, start_date, end_date)

    html = get_template("email/activity_report.html")

    context = {"profile": profile, "employee": employee}
    html_content = html.render(context)

    msg = EmailMultiAlternatives(
        _("Report"), strip_tags(html_content), settings.DEFAULT_FROM_EMAIL, [profile.user_account.email]
    )
    msg.attach_alternative(html_content, "text/html")
    title = _("all")
    if start_date and end_date:
        title = "{}_{}".format(start_date.strftime("%m/%d/%Y-%H:%M:%S"), end_date.strftime("%m/%d/%Y-%H:%M:%S"))
    msg.attach(
        "report_{}_{}.xlsx".format(employee.computed_str, title),
        report.read(),
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    )
    msg.send()


def create_report(employee, start_date, end_date):
    output = BytesIO()
    workbook = xlsxwriter.Workbook(output, {"in_memory": True})

    base_format = workbook.add_format({"font_name": "Liberation Sans", "font_size": 10})
    bold_format = workbook.add_format({"font_name": "Liberation Sans", "font_size": 10, "bold": True})
    related_activities = Activity.objects.filter(
        Q(participants__id=employee) | Q(creator__id=employee) | Q(assigned_to__id=employee)
    )
    activities = Activity.get_inrange_activities(related_activities, start_date, end_date)
    # HERE STARTS THE FIRST WORKSHEET
    for type in ActivityType.objects.all():
        width_cols = [5, 5, 16, 8, 7, 11, 12]
        activities_type = activities.filter(type=type.id)
        worksheet = workbook.add_worksheet(type.title)
        worksheet.write_string(0, 0, _("Title"), bold_format)
        worksheet.write_string(0, 1, _("Start"), bold_format)
        worksheet.write_string(0, 2, _("Duration (hours)"), bold_format)
        worksheet.write_string(0, 3, _("Location"), bold_format)
        worksheet.write_string(0, 4, _("Creator"), bold_format)
        worksheet.write_string(0, 5, _("Assigned to"), bold_format)
        worksheet.write_string(0, 6, _("Participants"), bold_format)

        worksheet.write_string(2, 0, _("Total"), bold_format)
        worksheet.write_string(2, 1, str(len(activities_type.all())), bold_format)

        for row, activity in enumerate(activities_type.all(), start=4):
            hours = (activity.period.upper - activity.period.lower).total_seconds() / 3600
            duration = format(hours, ".2f")
            creator = activity.creator
            creator_name = ""
            if creator:
                creator_name = creator.computed_str
            assigned_to = activity.assigned_to
            assigned_to_name = ""
            if assigned_to:
                assigned_to_name = assigned_to.computed_str
            participants = ", ".join([participant.computed_str for participant in activity.participants.all()])

            set_cell(worksheet, row, 0, activity.title, width_cols, base_format)
            set_cell(worksheet, row, 1, f"{activity.period.lower:%d.%m.%Y}", width_cols, base_format)
            set_cell(worksheet, row, 2, duration, width_cols, base_format)
            set_cell(worksheet, row, 2, activity.location, width_cols, base_format)
            set_cell(worksheet, row, 4, creator_name, width_cols, base_format)
            set_cell(worksheet, row, 5, assigned_to_name, width_cols, base_format)
            set_cell(worksheet, row, 6, participants, width_cols, base_format)

        for col, max_width in enumerate(width_cols):
            if max_width > 300:
                max_width = 300
            worksheet.set_column(col, col, max_width)
    workbook.close()
    output.seek(0)
    return output



from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Entry

from wbcrm.models.accounts import Account, AccountRole


class AccountFilter(wb_filters.FilterSet):
    parent = wb_filters.ModelChoiceFilter(
        label="Parent",
        queryset=Account.objects.all(),
        endpoint=Account.get_representation_endpoint(),
        value_key=Account.get_representation_value_key(),
        label_key=Account.get_representation_label_key(),
        hidden=True,
    )
    parent__isnull = wb_filters.BooleanFilter(field_name="parent", lookup_expr="isnull", hidden=True)
    customer = wb_filters.ModelChoiceFilter(
        label="Customer",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        method="filter_customer",
    )

    owner = wb_filters.ModelMultipleChoiceFilter(
        label="Owner",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
    )

    not_owner = wb_filters.ModelMultipleChoiceFilter(
        column_field_name="owner",
        lookup_icon="!=",
        lookup_label="Not Equals",
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
        method="filter_not_owner",
    )

    status = wb_filters.MultipleChoiceFilter(
        choices=Account.Status.choices, label="Status", initial=[Account.Status.OPEN.value]
    )

    def filter_not_owner(self, queryset, name, value):
        if value:
            return queryset.exclude(owner__in=value)
        return queryset

    def filter_customer(self, queryset, name, value):
        if value:
            return queryset.filter(id__in=Account.get_accounts_for_customer(value))
        return queryset

    class Meta:
        model = Account
        fields = {
            "reference_id": ["icontains"],
            "is_active": ["exact"],
            "parent": ["exact", "isnull"],
            "is_terminal_account": ["exact"],
            "is_public": ["exact"],
        }


class AccountRoleFilterSet(wb_filters.FilterSet):
    is_currently_valid = wb_filters.BooleanFilter(label="Valid", initial=True, field_name="is_currently_valid")

    class Meta:
        model = AccountRole
        fields = {
            "role_type": ["exact"],
            "entry": ["exact"],
            "is_hidden": ["exact"],
            "authorized_hidden_users": ["exact"],
        }



import zoneinfo
from contextlib import suppress
from datetime import date, datetime, timedelta
from typing import Any

import arrow
import numpy as np
from celery import shared_task
from dateutil.rrule import rrulestr
from django.conf import settings
from django.db import models, transaction
from django.db.models import Exists, OuterRef, Q
from django.db.models.query import QuerySet
from django.db.models.signals import m2m_changed, post_delete, post_save
from django.dispatch import receiver
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.translation import gettext, pgettext_lazy
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from dynamic_preferences.registries import global_preferences_registry
from ics.alarm import DisplayAlarm
from psycopg.types.range import TimestamptzRange
from slugify import slugify
from wbcore.contrib import workflow
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.agenda.signals import draggable_calendar_item_ids
from wbcore.contrib.ai.llm.decorators import llm
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import (
    Company,
    EmployerEmployeeRelationship,
    Entry,
    Person,
)
from wbcore.contrib.directory.preferences import get_main_company
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models import WBModel
from wbcore.utils.models import (
    CalendarItemTypeMixin,
    ComplexToStringMixin,
    DefaultMixin,
)

from wbcrm.models.llm.activity_summaries import analyze_activity
from wbcrm.models.recurrence import Recurrence
from wbcrm.synchronization.activity.shortcuts import get_backend
from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO


class DisplayAlarm2(DisplayAlarm):
    def __hash__(self):
        return hash(repr(self))


class ActivityType(DefaultMixin, CalendarItemTypeMixin, WBModel):
    class Score(models.TextChoices):
        HIGH = "4.0"
        MEDIUM = "3.0"
        LOW = "2.0"
        NONE = "0.0"
        MAX = "9.0"

    title = models.CharField(
        max_length=128,
        verbose_name=_("Title"),
        unique=True,
        blank=False,
        null=False,
    )

    slugify_title = models.CharField(
        max_length=128,
        unique=True,
        verbose_name="Slugified Title",
        blank=True,
        null=True,
    )

    score = models.CharField(
        max_length=8,
        verbose_name=_("Activity Heat Multiplier"),
        choices=Score.choices[:4],
        default=Score.LOW.value,
        unique=False,
        blank=False,
        null=False,
        help_text=_(
            "Used for the activity heat calculation. Multipliers range from low (i.e. e-mail) to medium (i.e. call) and high (i.e. meeting)."
        ),
    )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcrm:activitytype"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcrm:activitytyperepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"

    def __str__(self) -> str:
        return f"{self.title}"

    @classmethod
    def get_default_activity_type(cls):
        with suppress(cls.DoesNotExist):
            return cls.objects.get(default=True)

    class Meta:
        verbose_name = _("Activity Type")
        verbose_name_plural = _("Activity Types")

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)


@receiver(post_save, sender=ActivityType)
def trigger_activity_save(sender, instance: ActivityType, created: bool, raw: bool, **kwargs):
    """
    We need to trigger all activities' save methods to update their color and icon
    """

    if not raw:
        for activity in instance.activity.all():
            activity.save()


def has_permissions(instance, user):  # type: ignore
    if user.has_perm("wbcrm.change_activity"):
        if instance.visibility == CalendarItem.Visibility.PRIVATE:
            return instance.is_private_for_user(user)
        elif instance.visibility == CalendarItem.Visibility.CONFIDENTIAL:
            return instance.is_confidential_for_user(user)
        else:
            return True
    else:
        return False


@llm([analyze_activity])
@workflow.register(serializer_class="wbcrm.serializers.ActivityModelSerializer")
class Activity(Recurrence):
    summary = models.TextField(default="", blank=True, verbose_name=_("LLM Summary"))
    heat = models.PositiveIntegerField(null=True, blank=True)
    online_meeting = models.BooleanField(
        default=False,
        verbose_name=_("Online Meeting"),
        help_text=_("Check this if it happens online"),
    )

    class Status(models.TextChoices):
        CANCELLED = "CANCELLED", _("Cancelled")
        PLANNED = "PLANNED", _("Planned")
        FINISHED = "FINISHED", _("Finished")
        REVIEWED = "REVIEWED", _("Reviewed")

        @classmethod
        def get_color_map(cls):
            colors = [
                WBColor.RED_LIGHT.value,
                WBColor.YELLOW_LIGHT.value,
                WBColor.BLUE_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
            ]
            return [choice for choice in zip(cls, colors, strict=False)]

    class Importance(models.TextChoices):
        LOW = "LOW", _("Low")
        MEDIUM = "MEDIUM", _("Medium")
        HIGH = "HIGH", _("High")

    class ReminderChoice(models.TextChoices):
        NEVER = "NEVER", _("Never")
        EVENT_TIME = "EVENT_TIME", _("At time of event")
        MINUTES_5 = "MINUTES_5", _("5 minutes before")
        MINUTES_15 = "MINUTES_15", _("15 minutes before")
        MINUTES_30 = "MINUTES_30", _("30 minutes before")
        HOURS_1 = "HOURS_1", _("1 hour before")
        HOURS_2 = "HOURS_2", _("2 hour before")
        HOURS_12 = "HOURS_12", _("12 hour before")
        WEEKS_1 = "WEEKS_1", _("1 week before")

        @classmethod
        def get_minutes_correspondance(cls, name):
            _map = {
                "NEVER": -1,
                "EVENT_TIME": 0,
                "MINUTES_5": 5,
                "MINUTES_15": 15,
                "MINUTES_30": 30,
                "HOURS_1": 60,
                "HOURS_2": 120,
                "HOURS_12": 720,
                "WEEKS_1": 10080,
            }

            return _map[name]

    class Meta:
        verbose_name = _("Activity")
        verbose_name_plural = _("Activities")
        notification_types = [
            create_notification_type(
                "wbcrm.activity.participant",
                gettext("Activity Participant"),
                gettext("User notification when addeded to an activity."),
            ),
            create_notification_type(
                "wbcrm.activity.reminder",
                gettext("Activity Reminder"),
                gettext("Sends a reminder that an activity is starting soon."),
            ),
            create_notification_type(
                "wbcrm.activity.finished",
                gettext("Finished Activity"),
                gettext("Notifies a user of a finished activity that can be reviewed."),
            ),
            create_notification_type(
                "wbcrm.activity.daily_summary",
                gettext("Daily Summary"),
                gettext("Sends out a daily summary of a user's upcoming day."),
                web=False,
                mobile=False,
                email=True,
                is_lock=True,
            ),
            create_notification_type(
                "wbcrm.activity_sync.admin",
                gettext("Activity Sync Irregularities"),
                gettext("Admin notification to inform about irregularities of the activity sync."),
            ),
        ]

    status = FSMField(default=Status.PLANNED, choices=Status.choices, verbose_name=_("Status"))

    @transition(
        field=status,
        source=[Status.PLANNED],
        target=Status.FINISHED,
        permission=has_permissions,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activity",),
                icon=WBIcon.CONFIRM.icon,
                key="finish",
                label=_("Finish"),
                action_label=_("Finish"),
                description_fields=_("Are you sure you want to finish this activity?"),
            )
        },
    )
    def finish(self, by=None, description=None, **kwargs):
        self.cancel_recurrence()

    def can_finish(self):
        errors = dict()

        if not self.period:
            errors["period"] = [_("In this status this has to be provided.")]

        return errors

    @transition(
        field=status,
        source=[Status.PLANNED, Status.FINISHED],
        target=Status.REVIEWED,
        permission=has_permissions,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activity",),
                icon=WBIcon.REVIEW.icon,
                key="review",
                label=pgettext_lazy("Transition button label for Reviews", "Review"),
                action_label=pgettext_lazy("Transition action label for Reviews", "Review"),
                description_fields="",
                instance_display=create_simple_display([["result"], ["participants"], ["companies"]]),
            )
        },
    )
    def review(self, by=None, description=None, **kwargs):
        self.cancel_recurrence()

    def can_review(self):
        errors = self.can_finish()
        if not self.result or self.result == "" or self.result == "<p></p>":
            errors["result"] = [_("When reviewing an activity a result has to be provided!")]

        missing = self._check_employer_employees_entered()

        if missing_companies := missing.get("missing_companies_by_participant", None):
            participants_with_missing_companies = [participant.computed_str for participant in missing_companies]
            errors["companies"] = [
                _(
                    "You need to enter an employer for: {persons}",
                ).format(persons=", ".join(participants_with_missing_companies))
            ]

        if missing_participants := missing.get("missing_participants_by_company", None):
            companies_with_missing_participants = [company.name for company in missing_participants]
            errors["participants"] = [
                _(
                    "You need to enter an employee for: {companies}",
                ).format(companies=", ".join(companies_with_missing_participants))
            ]
        return errors

    @transition(
        field=status,
        source=[Status.PLANNED],
        target=Status.CANCELLED,
        permission=has_permissions,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activity",),
                icon=WBIcon.REJECT.icon,
                key="cancel",
                label=_("Cancel"),
                action_label=_("Cancel"),
                description_fields=_("Are you sure you want to cancel this activity?"),
            )
        },
    )
    def cancel(self, by=None, description=None, **kwargs):
        self.cancel_recurrence()

    description = models.TextField(default="", blank=True, verbose_name=_("Description"))
    result = models.TextField(default="", blank=True, verbose_name=_("Review"))
    type = models.ForeignKey(
        "wbcrm.ActivityType",
        related_name="activity",
        on_delete=models.PROTECT,
        verbose_name=_("Type"),
    )
    importance = models.CharField(
        max_length=16, default=Importance.LOW, choices=Importance.choices, verbose_name=_("Importance")
    )

    start = models.DateTimeField(blank=True, null=True, verbose_name=_("Start"))
    end = models.DateTimeField(blank=True, null=True, verbose_name=_("End"))

    reminder_choice = models.CharField(
        max_length=16,
        default=ReminderChoice.MINUTES_15,
        choices=ReminderChoice.choices,
        verbose_name=_("Reminder"),
        help_text=_(
            "Sends a mail and system notification to all participating internal employees before the start of the activity."
        ),
    )
    location = models.CharField(
        max_length=2048, null=True, blank=True, verbose_name=_("Location")
    )  # we increase the max lenght to 2048 to accomodate meeting URL (ICS and outlook uses the location field to share meeting link)
    location_longitude = models.DecimalField(
        max_digits=9, decimal_places=6, null=True, blank=True, verbose_name=_("Longitude")
    )
    location_latitude = models.DecimalField(
        max_digits=9, decimal_places=6, null=True, blank=True, verbose_name=_("Latitude")
    )
    created = models.DateTimeField(auto_now_add=True, verbose_name=_("Created"))
    creator = models.ForeignKey(
        "directory.Person",
        related_name="activities_owned",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Creator"),
        help_text=_("The creator of this activity"),
    )
    latest_reviewer = models.ForeignKey(
        "directory.Person",
        related_name="activities_reviewed",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Latest Reviewer"),
        help_text=_("The latest person to review the activity"),
    )
    reviewed_at = models.DateTimeField(verbose_name=_("Reviewed at"), null=True, blank=True)
    edited = models.DateTimeField(auto_now=True, verbose_name=_("Edited"))
    assigned_to = models.ForeignKey(
        "directory.Person",
        related_name="activities",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Assigned to"),
        help_text=_("The person in charge of handling this activity"),
    )
    companies = models.ManyToManyField(
        "directory.Company",
        related_name="company_participates",
        blank=True,
        verbose_name=_("Participating Companies"),
        help_text=_("The list of companies other than the main company that participate in this activity"),
        through="wbcrm.ActivityCompanyThroughModel",
        through_fields=("activity", "company"),
    )

    participants = models.ManyToManyField(
        "directory.Person",
        related_name="participates",
        blank=True,
        verbose_name=_("Participating Persons"),
        help_text=_("The list of participants"),
        through="wbcrm.ActivityParticipant",
        through_fields=("activity", "participant"),
    )

    groups = models.ManyToManyField(
        "wbcrm.Group",
        related_name="activities_for_group",
        blank=True,
        verbose_name=_("Groups"),
        help_text=_("Each member of the group will be added to the list of participants and companies automatically."),
    )
    preceded_by = models.ForeignKey(
        "self",
        related_name="followed_by",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Preceded by"),
        help_text=_("The preceding activity"),
    )
    disable_participant_check = models.BooleanField(
        default=False,
        verbose_name=_("Without Participating Company"),
    )
    metadata = models.JSONField(default=dict, blank=True)

    def __str__(self):
        return "%s" % (self.title,)

    def is_private_for_user(self, user) -> bool:
        return (
            self.visibility == CalendarItem.Visibility.PRIVATE
            and user.profile not in self.participants.all()
            and user.profile != self.assigned_to
            and user.profile != self.creator
        )

    def is_confidential_for_user(self, user) -> bool:
        return (
            self.visibility == CalendarItem.Visibility.CONFIDENTIAL
            and not CalendarItem.has_user_administrate_permission(user)
        )

    def get_extra_ics_kwargs(self) -> dict[str, Any]:
        res = {}
        res["created"] = arrow.get(self.created)
        if self.location:
            res["location"] = "".join(self.location)
        if self.description:
            res["description"] = self.description
        if self.period and self.reminder_choice:
            reminder = self.period.lower - timedelta(
                minutes=Activity.ReminderChoice.get_minutes_correspondance(self.reminder_choice)
            )
            a = DisplayAlarm2(trigger=reminder)
            res["alarms"] = set([a])
        return res

    def get_color(self) -> str:
        return self.type.color

    def get_icon(self) -> str:
        return self.type.icon

    def save(self, synchronize: bool = True, *args, **kwargs):
        pre_save_activity_dto = (
            Activity.all_objects.get(id=self.id)._build_dto() if self.id else None
        )  # we need to refetch to pre save activity from the database because self already contains the updated fields

        # Set reviewed
        if (
            self.status not in [Activity.Status.REVIEWED, Activity.Status.CANCELLED]
            and self.result
            and self.result not in ["<p></p>", ""]
        ):
            self.status = Activity.Status.REVIEWED

        if not self.period and self.start and self.end:
            if self.start == self.end:
                self.end = self.end + timedelta(seconds=1)
            self.period = TimestamptzRange(self.start, self.end)  # type: ignore

        if not (self.period or self.start or self.end):
            self.period = TimestamptzRange(timezone.now(), timezone.now() + timedelta(hours=1))

        if self.period:
            self.start, self.end = self.period.lower, self.period.upper  # type: ignore

        # If all day activity, we ensure period spans the full range
        if self.all_day and self.period:
            tz = zoneinfo.ZoneInfo(settings.TIME_ZONE)

            self.period = TimestamptzRange(
                lower=self.period.lower.astimezone(tz).replace(hour=0, minute=0, second=0),
                upper=self.period.upper.astimezone(tz).replace(hour=23, minute=59, second=59),
            )  # type
        self.is_cancelled = self.status == self.Status.CANCELLED

        # Logic to be called after a save happens (e.g synchronization). We get the activity DTO before saving that we passed around in the signal
        super().save(*args, **kwargs)

        if synchronize and self.is_active:
            if not self.is_recurrent or not (
                self.is_recurrent
                and self.parent_occurrence
                and (self.parent_occurrence.propagate_for_all_children or not pre_save_activity_dto)
            ):
                # countdown of at least 20 seconds is necessary to get m2m
                post_save_callback.apply_async(
                    (self.id,), {"pre_save_activity_dto": pre_save_activity_dto}, countdown=20
                )

        if not self.type:
            self.type = ActivityType.get_default_activity_type()

    def delete(self, synchronize: bool = True, **kwargs):
        # Logic to be called after a deletion happens (e.g synchronization). We get the activity DTO before deletion that we passed around in the signal
        if synchronize and Activity.objects.filter(id=self.id).exists():
            pre_delete_activity_dto = Activity.objects.get(id=self.id)._build_dto()
            super().delete(**kwargs)
            post_delete_callback.apply_async(
                (self.id,), {"pre_delete_activity_dto": pre_delete_activity_dto}, countdown=1
            )
        else:
            super().delete(**kwargs)

    def get_participants(self) -> QuerySet[Person]:
        """
        Get all participants for that activity.

        Returns:
            Queryset<Person> The participants
        """
        return Person.objects.filter(Q(participates__id=self.id)).distinct()

    def get_companies(self):
        """
        Get all companies for that activity.

        Returns:
            Queryset<Company> The companies participating in the activity
        """
        return Company.objects.filter(Q(company_participates__id=self.id)).distinct()

    def _check_employer_employees_entered(self) -> dict:
        if not self.disable_participant_check:
            participants = self.participants.all()
            companies = self.companies.all()

            missing_employers_for_participant = set()
            missing_employees_for_company = set()

            for participant in participants.exclude(
                Q(id__in=Person.objects.filter_only_internal()) | Q(employers__isnull=True)
            ):
                if not participant.employers.filter(id__in=companies).exists():
                    missing_employers_for_participant.add(participant)

            for company in companies.exclude(employees__isnull=True):
                if not company.employees.filter(id__in=participants).exists():
                    missing_employees_for_company.add(company)

            return {
                "missing_employers_for_participant": missing_employers_for_participant,
                "missing_employees_for_company": missing_employees_for_company,
            }
        return {}

    def participants_company_check_message(self) -> str:
        """Checks if the companies and participants fields have been filled in correctly.
            A warning is generated if employees or employers are missing in the corresponding field.

        Returns:
            str: The warning string
        """

        missing = self._check_employer_employees_entered()
        message = ""

        if missing.get("missing_employers_for_participant"):
            participants_with_missing_companies = [
                participant.computed_str for participant in missing["missing_employers_for_participant"]
            ]
            message += _("For the following participants you did not supply an employer: {persons}<br />").format(
                persons=", ".join(participants_with_missing_companies)
            )

        if missing.get("missing_employees_for_company"):
            companies_with_missing_participants = [
                company.computed_str for company in missing["missing_employees_for_company"]
            ]
            message += _("For the following companies you did not supply an employee: {companies}<br />").format(
                companies=", ".join(companies_with_missing_participants)
            )

        return message

    def get_occurrance_dates(self) -> list:
        """
        Returns a list with datetime values based on the recurrence options of an activity.

        :return:    list with datetime values
        :rtype:     list
        """

        occurrance_dates = []
        # dd = self.start.date()
        if not self.period:
            raise AttributeError(_("Period needs to be set for recurrence to work!"))
        dd = self.period.lower.date()
        # weekday = calendar.day_abbr[dd.weekday()].upper()[:2]
        # weekday_position = (dd.day + 6) // 7
        # weekdaycounters = collections.Counter(
        #     [calendar.weekday(dd.year, dd.month, d) for d in range(1, calendar.monthrange(dd.year, dd.month)[1] + 1)]
        # )
        # total_number_of_weekday = weekdaycounters[dd.weekday()]
        repeat_rule = self.repeat_choice
        # if self.repeat_choice == Recurrence.ReoccuranceChoice.MONTHLY_WEEKDAY:
        #     repeat_rule = f"RRULE:FREQ=MONTHLY;BYDAY={weekday};BYSETPOS={weekday_position}"
        # elif self.repeat_choice == Recurrence.ReoccuranceChoice.MONTHLY_LASTWEEKDAY:
        #     repeat_rule = (
        #         f"RRULE:FREQ=MONTHLY;BYDAY={weekday};BYSETPOS={weekday_position - total_number_of_weekday - 1}"
        #     )
        end_date = (
            self.recurrence_end + timedelta(days=1)
            if self.recurrence_end
            else global_preferences_registry.manager()["wbcrm__recurrence_activity_end_date"]
        )
        dstart = datetime.combine(dd, self.start.astimezone().time())
        if self.recurrence_count:
            occurrance_dates = list(rrulestr(repeat_rule + f";COUNT={self.recurrence_count+1}", dtstart=dstart))[1:]
        else:
            occurrance_dates = list(
                rrulestr(
                    repeat_rule + f";UNTIL={end_date.strftime('%Y%m%d')}",
                    dtstart=dstart.replace(tzinfo=None),
                )
            )[1:]

        return occurrance_dates

    def update_last_event(self):
        """
        Updates the entries last activity
        """

        activities = (
            Activity.objects.filter(visibility=CalendarItem.Visibility.PUBLIC)
            .filter(Q(companies__id=self.id) | Q(participants__id=self.id))
            .filter(
                (
                    Q(period__endswith__lt=timezone.now())
                    & Q(status__in=[Activity.Status.FINISHED, Activity.Status.REVIEWED])
                )
                | Q(status=Activity.Status.REVIEWED)
            )
        )
        if activities.exists() and (last_event := activities.latest("period__endswith")):
            self.last_event_id = last_event.id
            self.save()

    def get_participants_for_employer(self, employer: Entry) -> QuerySet[Person]:
        rels = (
            ActivityParticipant.objects.filter(activity=self)
            .annotate(
                is_employee=Exists(
                    EmployerEmployeeRelationship.objects.filter(
                        employee=OuterRef("participant"), employer=employer, primary=True
                    )
                )
            )
            .filter(is_employee=True)
        )
        return Person.objects.filter(id__in=rels.values("participant"))

    @staticmethod
    def get_inrange_activities(
        queryset: QuerySet["Activity"], start_date: date, end_date: date
    ) -> QuerySet["Activity"]:
        """
        Returns all activities taking place during the given interval. Accounts for the recurring activities as well.

        Args:
            queryset (Queryset[Activity]): The base queryset
            start_date (date): The starting point of the interval
            end_date (date): The end point of the interval

        Returns:
            queryset (Queryset[Activity]): A queryset of all activities with occurrences in the specified period
        """
        interval = TimestamptzRange(start_date, end_date)  # type: ignore
        return queryset.filter(period__overlap=interval)

    @staticmethod
    def get_companies_activities(queryset, value):
        """
        Return the activities whose companies are value.

        Arguments:
            queryset {Queryset<Activity>} -- The base queryset
            value {list<Entry>} -- A list of entries considered as companies
        Returns:
            queryset {Queryset<Activity>} -- A queryset whose companies includes value
        """
        return queryset.filter(companies__in=value).distinct()

    @classmethod
    def get_activities_for_user(cls, user, base_qs=None):
        if base_qs is None:
            base_qs = Activity.objects
        if user.is_superuser or user.profile.is_internal:
            queryset = base_qs.all()
        else:
            queryset = base_qs.filter(
                Q(creator=user.profile)
                | Q(assigned_to=user.profile)
                | Q(activity_participants__participant_id=user.profile.id)
                | Q(activity_participants__participant__in=user.profile.clients.all())
                | Q(companies__in=user.profile.clients.all())
            )

        return queryset.distinct()

    # Overriden Function from the recurrence framework
    def _handle_recurrence_m2m_forwarding(self, child):
        child.groups.set(self.groups.union(child.groups.all()))
        child.participants.set(self.participants.union(child.participants.all()))
        child.companies.set(self.companies.union(child.companies.all()))

    def does_recurrence_need_cancellation(self):
        return self.status in [
            Activity.Status.FINISHED,
            Activity.Status.REVIEWED,
            Activity.Status.CANCELLED,
        ]

    def get_recurrent_valid_children(self):
        return super().get_recurrent_valid_children().filter(status=Activity.Status.PLANNED)

    def _create_recurrence_child(self, start_datetime: datetime):
        child = Activity(
            assigned_to=self.assigned_to,
            all_day=self.all_day,
            conference_room=self.conference_room,
            creator=self.creator,
            description=self.description,
            disable_participant_check=self.disable_participant_check,
            importance=self.importance,
            visibility=self.visibility,
            location=self.location,
            location_longitude=self.location_longitude,
            location_latitude=self.location_latitude,
            parent_occurrence=self,
            period=TimestamptzRange(start_datetime, (start_datetime + self.duration)),
            recurrence_end=self.recurrence_end,
            recurrence_count=self.recurrence_count,
            reminder_choice=self.reminder_choice,
            repeat_choice=self.repeat_choice,
            title=self.title,
            type=self.type,
        )
        child.save(synchronize=False)
        return child

    def _build_dto(self):
        return ActivityDTO(
            metadata=self.metadata,
            title=self.title,
            period=self.period,
            description=self.description,
            participants=[
                ParticipantStatusDTO(
                    status=rel.participation_status,
                    status_changed=rel.status_changed,
                    person=rel.participant._build_dto(),
                )
                for rel in self.activity_participants.all()
            ],
            creator=self.creator._build_dto() if self.creator else None,
            visibility=self.visibility,
            reminder_choice=self.reminder_choice,
            is_cancelled=self.is_cancelled,
            all_day=self.all_day,
            online_meeting=self.online_meeting,
            location=self.location,
            conference_room=self.conference_room._build_dto() if self.conference_room else None,
            id=self.id,
            # parent_occurrence=self.parent_occurrence,
            # recurring_activities=self.get_recurrent_valid_children(),
            # invalid_recurring_activities=self.get_recurrent_invalid_children(),
            is_root=self.is_root,
            is_leaf=self.is_leaf,
            propagate_for_all_children=self.propagate_for_all_children,
            recurrence_end=self.recurrence_end,
            recurrence_count=self.recurrence_count,
            repeat_choice=self.repeat_choice,
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcrm:activity"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcrm:activityrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"


@receiver(m2m_changed, sender=Activity.participants.through)
def m2m_changed_participants(sender, instance: Activity, action: str, pk_set: set[int], **kwargs):
    """
    Handle the post custom logic when adding a participant. In that case, we call the relationship save method because we define a through model
    """
    if action in ["post_add", "pre_remove", "pre_clear"]:
        for participant_id in pk_set:
            rel = ActivityParticipant.objects.get(activity=instance, participant=participant_id)
            if action == "post_add":
                rel.save()
                if not instance.parent_occurrence:
                    send_employee_notification.delay(instance.id, participant_id)
            else:
                rel.delete()


@receiver(m2m_changed, sender=Activity.companies.through)
def m2m_changed_companies(sender, instance: Activity, action: str, pk_set: set[int], **kwargs):
    """
    Send a notification whenever a user who did not create the activity is added as a participant.
    """
    if pk_set:
        if "add" in action:
            if (main_company := get_main_company()) and main_company.id in pk_set:
                pk_set.remove(main_company.id)
        if "post" in action:
            for company_id in pk_set:
                entry = Entry.all_objects.get(id=company_id)
                if action == "post_add":
                    instance.entities.add(entry)
                elif not instance.get_participants_for_employer(entry).exists():
                    instance.entities.remove(entry)
        if action == "post_add":
            for company_id in pk_set:
                with suppress(
                    ActivityCompanyThroughModel.DoesNotExist
                ):  # we save to trigger the computed str computation. I don't know of any other choice as django only allow bulk create on m2m insertion
                    ActivityCompanyThroughModel.objects.get(
                        company_id=company_id,
                        activity=instance,
                    ).save()


@receiver(m2m_changed, sender=Activity.groups.through)  # type: ignore
def m2m_changed_groups(sender, instance: Activity, action, pk_set, **kwargs):
    from wbcrm.models.groups import Group

    if action == "post_add" or action == "post_remove" and pk_set:
        instance_participants = instance.participants.all()
        instance_companies = instance.companies.all()
        edited_groups = Group.objects.filter(id__in=pk_set)
        edited_groups_members = edited_groups.values_list("members__id", flat=True)
        edited_persons = Person.objects.filter(id__in=edited_groups_members)
        edited_companies = Company.objects.filter(id__in=edited_groups_members)

        if action == "post_add":
            instance.participants.set(instance_participants.union(edited_persons))
            instance.companies.set(instance_companies.union(edited_companies))
        else:
            # Get group members who are members in groups that are to be removed and are also members in groups that remain in the instance.
            # These group members are not to be removed from the instance.
            remaining_groups_members = instance.groups.exclude(id__in=pk_set).values_list("members__id", flat=True)
            members_in_different_groups = np.intersect1d(edited_groups_members, remaining_groups_members)  # type: ignore

            instance.participants.set(
                instance_participants.difference(edited_persons.exclude(id__in=members_in_different_groups))
            )
            instance.companies.set(
                instance_companies.difference(edited_companies.exclude(id__in=members_in_different_groups))
            )


@shared_task()
def post_save_callback(activity_id: int, pre_save_activity_dto: ActivityDTO = None):
    with suppress(Activity.DoesNotExist):
        activity = Activity.all_objects.get(id=activity_id)
        # Set calendar item entities once all activity m2m relations are settled
        activity_dto = activity._build_dto()

        if activity.is_recurrent:
            if activity.is_root and not pre_save_activity_dto:  # create the recurring activities.
                activity.generate_occurrences()
            elif activity.propagate_for_all_children:  # update occurrences from activity
                period_time_changed = False
                if (
                    activity_dto.period.lower.time() != pre_save_activity_dto.period.lower.time()
                    or activity_dto.period.upper.time() != pre_save_activity_dto.period.upper.time()
                ):
                    period_time_changed = True
                activity.forward_change(period_time_changed=period_time_changed)

            activity_dto.recurring_activities = [
                instance._build_dto() for instance in activity.get_recurrent_valid_children()
            ]
            activity_dto.invalid_recurring_activities = [
                instance._build_dto() for instance in activity.get_recurrent_invalid_children()
            ]

        if controller := get_backend():
            controller.handle_outbound(activity_dto, old_activity_dto=pre_save_activity_dto)


@shared_task()
def post_delete_callback(activity_id: int, pre_delete_activity_dto: ActivityDTO):
    if controller := get_backend():
        with suppress(Activity.DoesNotExist):
            activity_dto = Activity.all_objects.get(id=activity_id)._build_dto()
            controller.handle_outbound(activity_dto, old_activity_dto=pre_delete_activity_dto, is_deleted=True)


@shared_task
def send_employee_notification(activity_id: int, participant_id):
    """Sends all employees that were added to the activity as participants a notification via system and mail

    Args:
        activity_id: The activity instance id to which participants were added
        participant_id: id of the participant to send notification to
    """
    with suppress(Activity.DoesNotExist):
        activity = Activity.all_objects.get(id=activity_id)
        with suppress(Person.DoesNotExist):
            employee = Person.objects.filter_only_internal().get(id=participant_id)
            if activity.creator != employee:
                activity_type_label = activity.type.title.lower()
                description = (
                    activity.description if activity.description and activity.description != "<p></p>" else None
                )
                message = render_to_string(
                    "email/activity.html",
                    {
                        "participants": activity.participants.all(),
                        "type": activity_type_label,
                        "title": activity.title,
                        "start": activity.period.lower if activity.period else "",
                        "end": activity.period.upper if activity.period else "",
                        "description": description,
                    },
                )
                if activity.period:
                    start_datetime: datetime = activity.period.lower
                    datetime_string = _(" starting at the {} at {}").format(
                        start_datetime.strftime("%d.%m.%Y"), start_datetime.strftime("%H:%M:%S")
                    )
                else:
                    datetime_string = ""
                creator_string = (
                    _("{} added you").format(str(activity.creator))
                    if activity.creator
                    else _("You were automatically added")
                )

                send_notification(
                    code="wbcrm.activity.participant",
                    title=_("{} as participant in a {}{}").format(
                        creator_string, activity_type_label, datetime_string
                    ),
                    body=message,
                    user=employee.user_account,
                    reverse_name="wbcrm:activity-detail",
                    reverse_args=[activity.id],
                )


# /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#                                               >>> Activity Participants <<<
# /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


def is_participant(instance, user):
    return instance.participant == user.profile and instance.activity.period.lower >= timezone.now()


def conference_room_is_videoconference_capable(instance):
    return not instance.activity.conference_room or instance.activity.conference_room.is_videoconference_capable


class ActivityCompanyThroughModel(ComplexToStringMixin, models.Model):
    activity = models.ForeignKey(
        on_delete=models.CASCADE,
        to="wbcrm.Activity",
        verbose_name=_("Activity"),
        related_name="activity_companies",
    )
    company = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Company",
        verbose_name=_("Company"),
        related_name="activity_companies",
    )
    customer_status = models.ForeignKey(
        to="directory.CustomerStatus",
        related_name="activity_companies",
        on_delete=models.SET_NULL,
        verbose_name=_("Initial Customer Status"),
        help_text=_("The Customer Status at activity creation time"),
        null=True,
        blank=True,
    )

    def __str__(self) -> str:
        return f"{self.activity} - {self.company} ({self.customer_status})"

    def save(self, *args, **kwargs):
        self.customer_status = self.company.customer_status
        super().save(*args, **kwargs)

    def compute_str(self):
        rep = self.company.computed_str
        if self.customer_status:
            rep += f" ({self.customer_status.title})"
        return rep


class ActivityParticipant(models.Model):
    class ParticipationStatus(models.TextChoices):
        CANCELLED = "CANCELLED", _("Cancelled")
        MAYBE = "MAYBE", _("Maybe")
        ATTENDS = "ATTENDS", _("Attends")
        NOTRESPONDED = "NOTRESPONDED", _("Not Responded")
        ATTENDS_DIGITALLY = "ATTENDS_DIGITALLY", _("Attends Digitally")
        PENDING_INVITATION = "PENDING_INVITATION", _("Pending Invitation")

    activity = models.ForeignKey(
        on_delete=models.CASCADE,
        to="wbcrm.Activity",
        verbose_name=_("Activity"),
        related_name="activity_participants",
    )
    participant = models.ForeignKey(
        on_delete=models.CASCADE,
        to="directory.Person",
        verbose_name=_("Participant"),
        related_name="activity_participants",
    )
    participation_status = FSMField(
        default=ParticipationStatus.PENDING_INVITATION,
        choices=ParticipationStatus.choices,
        verbose_name=_("Participation Status"),
    )
    status_changed = models.DateTimeField(auto_now=True)

    def __str__(self) -> str:
        return _("Status of {participant} for activity {title} is: {status}").format(
            participant=self.participant.computed_str,
            title=self.activity.title,
            status=self.ParticipationStatus[self.participation_status].label,
        )

    class Meta:
        constraints = [models.UniqueConstraint(name="unique_participant", fields=["activity", "participant"])]
        verbose_name = _("Activity's Participant")
        verbose_name_plural = _("Activities' Participants")

    def save(self, *args, **kwargs):
        self.activity.entities.add(self.participant)
        with suppress(EmployerEmployeeRelationship.DoesNotExist):
            rel = EmployerEmployeeRelationship.objects.get(employee=self.participant, primary=True)
            self.activity.companies.add(rel.employer)
        # Set the status 'Attends' by default for activity creator
        if self.activity.creator == self.participant:
            self.status = ActivityParticipant.ParticipationStatus.ATTENDS

        pre_save_participant_dto = ActivityParticipant.objects.get(id=self.id)._build_dto() if self.id else None
        # Logic to be called after a save happens (e.g synchronization). We get the activity Participant DTO before saving that we passed around in the signal
        super().save(*args, **kwargs)
        # we activate synchronization only if the rel was already created (we expect synchronization on the activity creation itself)
        post_save_participant_callback.apply_async(
            (self.id,), {"pre_save_participant_dto": pre_save_participant_dto}, countdown=10
        )

    def delete(self, *args, **kwargs):
        self.activity.entities.remove(self.participant.entry_ptr)
        with suppress(EmployerEmployeeRelationship.DoesNotExist):
            rel = EmployerEmployeeRelationship.objects.get(employee=self.participant, primary=True)
            # delete only if no other participants are of the same company
            if not self.activity.get_participants_for_employer(rel.employer).exclude(id=rel.employee.id).exists():
                self.activity.companies.remove(rel.employer)
                self.activity.entities.remove(rel.employer.entry_ptr)
        if self.activity.is_active:
            post_delete_participant_callback.apply_async((self.id,), countdown=10)
        super().delete(*args, **kwargs)

    def _build_dto(self):
        return ParticipantStatusDTO(
            person=self.participant._build_dto(),
            status_changed=self.status_changed,
            status=self.participation_status,
            activity=self.activity._build_dto(),
            id=self.id,
        )

    @transition(
        field=participation_status,
        source=[
            ParticipationStatus.PENDING_INVITATION,
            ParticipationStatus.NOTRESPONDED,
            ParticipationStatus.ATTENDS,
            ParticipationStatus.ATTENDS_DIGITALLY,
            ParticipationStatus.CANCELLED,
        ],
        target=ParticipationStatus.MAYBE,
        permission=is_participant,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activityparticipant",),
                icon=WBIcon.QUESTION.icon,
                key="maybe",
                label=_("Maybe"),
                action_label=_("Setting Maybe"),
            )
        },
    )
    def maybe(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=participation_status,
        source=[
            ParticipationStatus.PENDING_INVITATION,
            ParticipationStatus.NOTRESPONDED,
            ParticipationStatus.MAYBE,
            ParticipationStatus.ATTENDS_DIGITALLY,
            ParticipationStatus.CANCELLED,
        ],
        target=ParticipationStatus.ATTENDS,
        permission=is_participant,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activityparticipant",),
                icon=WBIcon.APPROVE.icon,
                key="attends",
                label=_("Accept"),
                action_label=_("Accepting"),
                description_fields=_("Are you sure you want to participate in this activity?"),
            )
        },
    )
    def attends(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=participation_status,
        source=[
            ParticipationStatus.PENDING_INVITATION,
            ParticipationStatus.NOTRESPONDED,
            ParticipationStatus.MAYBE,
            ParticipationStatus.ATTENDS,
            ParticipationStatus.CANCELLED,
        ],
        target=ParticipationStatus.ATTENDS_DIGITALLY,
        permission=is_participant,
        conditions=[conference_room_is_videoconference_capable],
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activityparticipant",),
                icon="laptop",
                key="attendsdigitally",
                label=_("Attend Digitally"),
                action_label=_("Setting Attendance"),
                description_fields=_("Are you sure you want to attend digitally in this activity?"),
            )
        },
    )
    def attendsdigitally(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=participation_status,
        source=[
            ParticipationStatus.PENDING_INVITATION,
            ParticipationStatus.NOTRESPONDED,
            ParticipationStatus.MAYBE,
            ParticipationStatus.ATTENDS,
            ParticipationStatus.ATTENDS_DIGITALLY,
        ],
        target=ParticipationStatus.CANCELLED,
        permission=is_participant,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activityparticipant",),
                icon=WBIcon.DENY.icon,
                key="cancelled",
                label=_("Decline"),
                action_label=_("Decline"),
                description_fields=_("Are you sure you want to decline to participate in this activity?"),
            )
        },
    )
    def cancelled(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=participation_status,
        source=ParticipationStatus.CANCELLED,
        target=ParticipationStatus.NOTRESPONDED,
        permission=lambda instance, user: instance.activity.creator == user.profile != instance.participant,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:activityparticipant",),
                icon=WBIcon.REDO.icon,
                key="notresponded",
                label=_("Resend Invitation"),
                action_label=_("Resending Invitation"),
                description_fields=_(
                    "Are you sure you want to send this person an invitation to participate in this activity again?"
                ),
            )
        },
    )
    def notresponded(self, by=None, description=None, **kwargs):
        pass

    @classmethod
    def get_representation_value_key(cls):
        return "id"


@receiver(post_save, sender=Activity)
def post_save_activity(sender, instance, created, raw, **kwargs):
    # need to the post save because instance might not be created yet in the save method
    if not raw and created:
        if instance.creator:
            instance.entities.add(instance.creator)
        if instance.assigned_to:
            instance.entities.add(instance.assigned_to)


@shared_task()
def post_save_participant_callback(
    activity_participant_id: int, pre_save_participant_dto: ParticipantStatusDTO = None
):
    if controller := get_backend():
        with suppress(ActivityParticipant.DoesNotExist):
            participant_dto = ActivityParticipant.objects.get(id=activity_participant_id)._build_dto()
            controller.handle_outbound_participant(participant_dto, old_participant_dto=pre_save_participant_dto)


@shared_task()
def post_delete_participant_callback(activity_participant_id: int):
    if controller := get_backend():
        with suppress(ActivityParticipant.DoesNotExist):
            participant_dto = ActivityParticipant.objects.get(id=activity_participant_id)._build_dto()
            controller.handle_outbound_participant(participant_dto, is_deleted=True)


@shared_task()
def send_invitation_participant_as_task(activity_id: int):
    if controller := get_backend():
        with suppress(Activity.DoesNotExist):
            activity_dto = Activity.all_objects.get(id=activity_id)._build_dto()
            participants_dto = [
                participant._build_dto()
                for participant in ActivityParticipant.objects.filter(
                    activity_id=activity_id,
                    participation_status=ActivityParticipant.ParticipationStatus.PENDING_INVITATION,
                )
            ]
            controller.handle_outbound_external_participants(activity_dto, participants_dto)


@receiver(draggable_calendar_item_ids, sender="agenda.CalendarItem")
def activity_draggable_calendar_item_ids(sender, request, **kwargs) -> QuerySet[CalendarItem]:
    return Activity.objects.filter(
        (Q(creator=request.user.profile) | Q(assigned_to=request.user.profile)) & Q(status=Activity.Status.PLANNED)
    ).values("id")


@receiver(post_save, sender=EmployerEmployeeRelationship)
def post_save_eer(sender, instance: EmployerEmployeeRelationship, created, raw, **kwargs):
    """
    Post save EER signal: Triggers the post_save signals of the employee which updates his computed_str and adds the
    employer to future planned activities if it became the only employer
    """

    if not raw and created and sender.objects.filter(employee=instance.employee, primary=True).count() == 1:
        transaction.on_commit(lambda: add_employer_to_activities.delay(instance.pk))


@receiver(post_delete, sender=EmployerEmployeeRelationship)
def post_delete_eer(sender, instance: EmployerEmployeeRelationship, **kwargs):
    """
    Post delete EER signal: Triggers the post_delete signals of the employee which updates his computed_str and adds the
    employer to future planned activities if it became the only employer
    """
    if sender.objects.filter(employee=instance.employee, primary=True).count() == 1:
        eer_obj = sender.objects.get(employee=instance.employee, primary=True)
        transaction.on_commit(lambda: add_employer_to_activities.delay(eer_obj.pk))


@shared_task
def add_employer_to_activities(eer_id: int):
    with suppress(EmployerEmployeeRelationship.DoesNotExist):
        eer_obj = EmployerEmployeeRelationship.objects.get(id=eer_id)

        for activity in Activity.objects.filter(
            status=Activity.Status.PLANNED,
            start__gte=timezone.now(),
            participants=eer_obj.employee,
        ):
            if eer_obj.employer not in activity.entities.all():
                activity.entities.add(eer_obj.employer)
            if eer_obj.employer not in activity.companies.all():
                activity.companies.add(eer_obj.employer)



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.directory.viewsets.display import CompanyModelDisplay
from wbcore.dispatch import receiver_all_subclasses
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import add_display_pages
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
    default,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ProductDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="is_competitor", label=_("Is Competitor")),
                dp.Field(key="prospects", label=_("Prospects")),
            )
        )

    def get_instance_display(self) -> Display:
        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["title", "is_competitor", "prospects"]],
                            grid_template_columns=[
                                "minmax(min-content, 0.5fr)",
                                "minmax(min-content, 1fr)",
                                "minmax(min-content, 1fr)",
                            ],
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(6),
                        ),
                    },
                ),
            ]
        )


class ProductCompanyRelationshipDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="product", label=_("Product")),
                dp.Field(key="competitor_product", label=_("Is Competitor")),
            )
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["product"]])


@receiver_all_subclasses(add_display_pages, sender=CompanyModelDisplay)
def add_interested_product_page(sender, *args, **kwargs):
    return [
        Page(
            title=_("Interests"),
            layouts={
                default(): Layout(
                    grid_template_areas=[["interested_products_section"]],
                    grid_auto_columns="minmax(min-content, 1fr)",
                    grid_auto_rows=Style.MIN_CONTENT,
                    sections=[
                        Section(
                            key="interested_products_section",
                            title=_("Interested Products"),
                            collapsible=False,
                            display=Display(
                                pages=[
                                    Page(
                                        title=_("Interested Products"),
                                        layouts={
                                            default(): Layout(
                                                grid_template_areas=[["interested_products_table"]],
                                                inlines=[
                                                    Inline(
                                                        key="interested_products_table",
                                                        endpoint="interested_products",
                                                    )
                                                ],
                                            )
                                        },
                                    ),
                                ]
                            ),
                        )
                    ],
                ),
            },
        )
    ]



from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import BooleanPreference, IntegerPreference

general = Section("wbcrm")


@global_preferences_registry.register
class CheckForMandatoryParticipants(BooleanPreference):
    section = general
    name = "mandatory_participants"
    default = True

    verbose_name = _("Check for mandatory participants")
    help_text = _(
        'Determines whether or not companies must be entered as participants in the "Companies" field for activities.'
    )


@global_preferences_registry.register
class RecurrenceActivityEndDate(IntegerPreference):
    section = general
    name = "recurrence_maximum_allowed_days"
    # this short default value, is just for the moment to test the process and until we know how to handle the recurring activities more efficient
    default = 10 * 365

    verbose_name = _("The default Maximum allowed days")


@global_preferences_registry.register
class RecurrenceActivityDateListLength(IntegerPreference):
    section = general
    name = "recurrence_maximum_count"
    # this short default value, is just for the moment to test the process and until we know how to handle the recurring activities more efficient
    default = 366

    # verbose_name = _("For each date in the date list we create a child activity.number at which the date list will be cut.")



from contextlib import suppress
from datetime import date
from typing import Any

from django.contrib.auth import get_user_model
from django.http import HttpRequest, HttpResponse
from django.utils.translation import gettext_lazy
from wbcore.contrib.notifications.dispatch import send_notification

from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO
from wbcrm.typings import User as UserDTO

User = get_user_model()


class SyncBackend:
    METADATA_KEY = None

    def open(self):
        """
        Allows to perform primary operations or to open a communication channel for synchronization,
        such as defining the necessary configurations to send requests
        """
        pass

    def close(self):
        """
        Close the communication channel and unset configuration
        """
        pass

    def _validation_response(self, request: HttpRequest) -> HttpResponse:
        """
        send a response to the external calendar if necessary to validate the endpoint
        """
        return None

    def _is_inbound_request_valid(self, request: HttpRequest) -> bool:
        """
        Valid function to ensure that the request received meets expectations
        """
        raise NotImplementedError

    def _get_events_from_request(self, request: HttpRequest) -> list[dict[str, Any]]:
        """
        list of events following the notification received
        """
        raise NotImplementedError

    def _deserialize(self, event: dict[str, Any]) -> tuple[ActivityDTO, bool, UserDTO]:
        """
        convert the dictionary received to a valid format of an activity
        """
        raise NotImplementedError()

    def _serialize(self, activity_dto: ActivityDTO, created: bool = False) -> dict[str, Any]:
        """
        convert activity data transfer object to event dictionary
        """
        raise NotImplementedError()

    def _stream_deletion(self, activity_dto: ActivityDTO):
        """
        allow the deletion of the event in the external calendar
        we use the event_id stored in activity_dto's metadata to retrieve the event
        """
        raise NotImplementedError()

    def _stream_creation(
        self, activity_dto: ActivityDTO, activity_dict: dict[str, Any]
    ) -> tuple[ActivityDTO, dict[str, Any]]:
        """
        allow the creation of the event in the external calendar
        param: activity_dict: dictionary used to create the event

        we return a tuple of activity, metadata which contains the external id to be store in the activity
        """
        raise NotImplementedError()

    def _stream_update(
        self,
        activity_dto: ActivityDTO,
        activity_dict: dict[str, Any],
        only_participants_changed: bool = False,
        external_participants: list | None = None,
        keep_external_description: bool = False,
    ) -> tuple[ActivityDTO, dict[str, Any]]:
        """
        allow to update the event in the external calendar
        param: activity_dict: dictionary used to update the event
               activity_dto: we use the metadata of the activity to retrieve the event
               only_participants_changed: boolean to know if only the participants need to be update
               external_participants: list of external participants, that must be added to the current list of participants to avoid their deletion when the activity is updated
               keep_external_description: boolean to know if the description must be deleted or not before the update of the event
        """
        raise NotImplementedError()

    def _stream_update_only_attendees(self, activity_dto: ActivityDTO, participants_dto: list[ParticipantStatusDTO]):
        """
        allow to update only the attendees of the event in the external calendar
        """
        raise NotImplementedError()

    def _stream_extension_event(self, activity_dto: ActivityDTO) -> None:
        """
        Extend external event with custom data
        this allows us for example to add additional information to the event to easily identify it for a recurring activities
        """
        pass

    def _stream_forward(self, activity_dto: ActivityDTO, participants_dto: list[ParticipantStatusDTO]):
        """
        allow to forward an event to a new participant. the external calendar
        send an invitation to all participants and avoid sending an update of the activity to all participants
        """
        raise NotImplementedError()

    def _stream_participant_change(
        self, participant_dto: ParticipantStatusDTO, is_deleted: bool = False, wait_before_changing: bool = False
    ):
        """
        allow to update the status of an event participant
        """
        raise NotImplementedError()

    def _set_web_hook(self, user: "User") -> dict[str, dict[str, Any]]:
        """
        allows to activate the webhook for a user
        returns a dictionary that will be stored in the metadata of the ser
        """
        raise NotImplementedError()

    def _stop_web_hook(self, user: "User") -> dict[str, dict[str, Any]]:
        """
        allows to strop the webhook for a user and deletes the data stored in the metadata
        """
        raise NotImplementedError()

    def _check_web_hook(self, user: "User") -> bool:
        """
        return a boolean to know if a subscription is activated or not for a user
        """
        raise NotImplementedError()

    def set_web_hook(self, user: "User"):
        """
        allows to be sure that the metadata are saved by specifying the backend type.
        """
        new_metadata = self._set_web_hook(user)
        user.metadata.setdefault(self.METADATA_KEY, {})
        user.metadata[self.METADATA_KEY] = new_metadata
        user.save()

    def stop_web_hook(self, user: "User"):
        new_metadata = self._stop_web_hook(user)
        user.metadata.setdefault(self.METADATA_KEY, {})
        user.metadata[self.METADATA_KEY] = new_metadata
        user.save()

    def check_web_hook(self, user: "User") -> bool:
        try:
            return self._check_web_hook(user)
        except NotImplementedError:
            return False

    def renew_web_hooks(self) -> None:
        """
        Allows to renew existing webhooks of all users
        """
        pass

    def _get_webhook_inconsistencies(self) -> str:
        """
        return a message of anomalies that will be notified to the administrator/persons set in the preferences
        """
        raise NotImplementedError()

    def notify_admins_of_webhook_inconsistencies(self, emails: list) -> None:
        """
        the purpose is to make sure that the authorized persons receive the messages in case a webhook has been deactivated or not renewed correctly.
        """
        with suppress(NotImplementedError):
            if emails and (message := self._get_webhook_inconsistencies()):
                for recipient in User.objects.filter(email__in=emails):
                    send_notification(
                        code="wbcrm.activity_sync.admin",
                        title=gettext_lazy("Notify admins of event webhook inconsistencies - {}").format(date.today()),
                        body=f"<ul>{message}</ul>",
                        user=recipient,
                    )

    def get_external_event(self, activity_dto: ActivityDTO) -> dict:
        """
        Get an event of external calendar.
        """
        pass

    def get_external_participants(
        self, activity_dto: ActivityDTO, internal_participants_dto: list[ParticipantStatusDTO]
    ) -> list[str, Any]:
        """
        Get external participants of an external event
        """
        return []

    def _is_participant_valid(self, user: "User") -> bool:
        return user.is_active and user.is_register

    def is_valid(self, activity: ActivityDTO) -> bool:
        # Synchronize only if the creator or at least one participant has an active subscription
        participants = [activity.creator.email] if activity.creator else []
        participants.extend(list(map(lambda x: x.person.email, activity.participants)))
        return any([self._is_participant_valid(user) for user in User.objects.filter(email__in=set(participants))])



from celery import shared_task
from django.db import models
from django.db.models.signals import m2m_changed, pre_delete
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.directory.models import Company, Entry, Person
from wbcore.models import WBModel

from wbcrm.models import Activity


class Group(WBModel):
    title = models.CharField(max_length=255, unique=True, verbose_name=_("Title"))
    members = models.ManyToManyField("directory.Entry", related_name="groups", verbose_name=_("Members"))

    class Meta:
        verbose_name = _("Group")
        verbose_name_plural = _("Groups")

    def __str__(self):
        return self.title

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcrm:group"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcrm:grouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"


@receiver(m2m_changed, sender=Group.members.through)
def m2m_changed_members(sender, instance, action, pk_set, **kwargs):
    """
    M2m changed Group signal: Change participants, companies and entities of future planned activities/calendar items
    if a group's members get updated
    """
    if action == "post_add":
        add_changed_group_members(instance, pk_set)

    if action == "post_remove":
        remove_changed_group_members(instance, pk_set)


@receiver(pre_delete, sender=Group)
def pre_delete_group(sender, instance, **kwargs):
    """
    Post delete Group signal: Remove members from future planned activities/calendar items if a group was deleted
    """
    remove_deleted_groups_members(instance)


@shared_task
def add_changed_group_members(instance, pk_set):
    for activity in Activity.objects.filter(
        status=Activity.Status.PLANNED, start__gte=timezone.now(), groups=instance
    ):
        item = CalendarItem.all_objects.get(id=activity.id)
        for member in pk_set:
            entry = Entry.all_objects.get(id=member)
            if entry not in item.entities.all():
                item.entities.add(entry)
            if entry.is_company:
                company = Company.all_objects.get(id=entry.id)
                if company not in activity.companies.all():
                    activity.companies.add(company)
            else:
                person = Person.all_objects.get(id=entry.id)
                if person not in activity.participants.all():
                    activity.participants.add(person)


@shared_task
def remove_changed_group_members(instance, pk_set):
    for activity in Activity.objects.filter(
        status=Activity.Status.PLANNED, start__gte=timezone.now(), groups=instance
    ):
        item = CalendarItem.all_objects.get(id=activity.id)
        for member in pk_set:
            entry = Entry.all_objects.get(id=member)
            if entry in item.entities.all():
                item.entities.remove(entry)
            if entry.is_company:
                company = Company.all_objects.get(id=entry.id)
                if company in activity.companies.all():
                    activity.companies.remove(company)
            else:
                person = Person.all_objects.get(id=entry.id)
                if person in activity.participants.all():
                    activity.participants.remove(person)


@shared_task
def remove_deleted_groups_members(instance):
    for activity in Activity.objects.filter(
        status=Activity.Status.PLANNED, start__gte=timezone.now(), groups=instance
    ):
        item = CalendarItem.all_objects.get(id=activity.id)
        for member in instance.members.all():
            if member in item.entities.all():
                item.entities.remove(member)
            if member.is_company:
                company = Company.all_objects.get(id=member.id)
                if company in activity.companies.all():
                    activity.companies.remove(company)
            else:
                person = Person.all_objects.get(id=member.id)
                if person in activity.participants.all():
                    activity.participants.remove(person)



import pytest
from django.db.models import Q
from django.forms.models import model_to_dict
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.messages import InMemoryMessageStorage

from wbcrm.models.activities import Activity
from wbcrm.models.recurrence import Recurrence
from wbcrm.preferences import (
    get_maximum_allowed_recurrent_date,
    get_recurrence_maximum_count,
)
from wbcrm.serializers import ActivityModelSerializer
from wbcrm.viewsets import ActivityViewSet


@pytest.mark.django_db
class RecurrenceFixture:
    @pytest.fixture
    def occurrence_fixture(self, activity_factory):
        return activity_factory(repeat_choice=Recurrence.ReoccuranceChoice.DAILY, recurrence_count=3, preceded_by=None)

    @pytest.fixture
    def occurrences_fixture(self, occurrence_fixture):
        return Activity.objects.filter(Q(id=occurrence_fixture.id) | Q(parent_occurrence=occurrence_fixture)).order_by(
            "start"
        )


@pytest.mark.django_db
class TestRecurrenceModel(RecurrenceFixture):
    @pytest.mark.parametrize("repeat_choice", [*list(Recurrence.ReoccuranceChoice)])
    def test_is_recurrent(self, repeat_choice, recurring_activity_factory):
        act = recurring_activity_factory(repeat_choice=repeat_choice)
        assert act.is_recurrent == (act.repeat_choice != Recurrence.ReoccuranceChoice.NEVER)

    def test_recurring_activity_without_end(self, recurring_activity_factory):
        act = recurring_activity_factory(repeat_choice=Recurrence.ReoccuranceChoice.DAILY, recurrence_count=None)
        occurrences = Activity.objects.filter(parent_occurrence=act)
        assert occurrences.count() <= get_recurrence_maximum_count()
        assert (
            occurrences.order_by("period__startswith").last().period.lower.date()
            <= get_maximum_allowed_recurrent_date()
        )

    def test_is_root(self, occurrences_fixture):
        for occurrence in occurrences_fixture:
            assert occurrence.is_root == (occurrence.is_recurrent and not occurrence.parent_occurrence)

    def test_is_leaf(self, occurrences_fixture):
        for occurrence in occurrences_fixture:
            assert occurrence.is_leaf == (occurrence.is_recurrent and not occurrence.next_occurrence)

    def test_next_occurrence(self, occurrences_fixture):
        last_row = len(occurrences_fixture) - 1
        for row, occurrence in enumerate(occurrences_fixture):
            if row == last_row:
                assert occurrence.next_occurrence is None
            else:
                assert occurrence.next_occurrence == occurrences_fixture[row + 1]

    def test_previous_occurrence(self, occurrences_fixture):
        for row, occurrence in enumerate(occurrences_fixture):
            if row == 0:
                assert occurrence.previous_occurrence is None
            else:
                assert occurrence.previous_occurrence == occurrences_fixture[row - 1]

    # def test_get_recurrent_valid_children(self, occurrences_fixture):
    #     parent = occurrences_fixture[0]
    #     assert parent.get_recurrent_valid_children().count() == len(occurrences_fixture[1:])
    #     assert parent.get_recurrent_invalid_children().count() == 0
    #     occurrences_fixture[1].cancel()
    #     assert parent.get_recurrent_valid_children().count() == len(occurrences_fixture[1:]) - 1
    #     assert parent.get_recurrent_invalid_children().count() == 1
    #     occurrences_fixture[2].review()
    #     assert parent.get_recurrent_valid_children().count() == len(occurrences_fixture[1:]) - 2
    #     assert parent.get_recurrent_invalid_children().count() == 2
    #     occurrences_fixture[3].finish()
    #     assert parent.get_recurrent_valid_children().count() == len(occurrences_fixture[1:]) - 3
    #     assert parent.get_recurrent_invalid_children().count() == 3

    @pytest.mark.parametrize("position, only_next_occurrence", [(0, False), (0, True), (1, False), (1, True)])
    def test_forward_deletion(self, occurrences_fixture, only_next_occurrence, position):
        nb_activities = Activity.objects.count()
        occ = occurrences_fixture[position]
        next_occurrences = occurrences_fixture.filter(period__startswith__gt=occ.period.lower)
        next_occ = next_occurrences.first()
        nb_next_occurrences = next_occurrences.count()
        assert nb_activities == 4
        if only_next_occurrence:
            occ.forward_deletion([next_occ.id])
            assert Activity.objects.count() == nb_activities - 1 == 3
        else:
            occ.forward_deletion()
            assert Activity.objects.count() == nb_activities - nb_next_occurrences

    def test_claim_parent_hood(self, occurrences_fixture, activity_factory):
        single_activity = activity_factory(repeat_choice=Recurrence.ReoccuranceChoice.NEVER, preceded_by=None)
        single_activity.claim_parent_hood()
        assert single_activity.is_root is False

        for occurrence in occurrences_fixture:
            occurrence.claim_parent_hood()
            if occurrence.is_leaf:
                assert occurrence.is_root is False
            else:
                assert occurrence.is_root is True

    @pytest.mark.parametrize("position, propagation", [(0, False), (0, True), (1, False), (1, True)])
    def test_delete(self, position, propagation, occurrences_fixture):
        occ = occurrences_fixture[position]
        next_occurrences = occurrences_fixture.filter(period__startswith__gt=occ.period.lower)
        next_occ = next_occurrences.first()
        nb_next_occurrences = next_occurrences.count()
        assert Activity.objects.count() == len(occurrences_fixture) == 4
        assert nb_next_occurrences == 3 - position
        Activity.objects.filter(id=occ.id).update(propagate_for_all_children=propagation)
        occ.refresh_from_db()
        was_root = True if occ.is_root else False
        occ.delete()
        if propagation:
            assert Activity.objects.count() == 3 - nb_next_occurrences
        else:
            assert Activity.objects.count() == 3
            next_occ.refresh_from_db()
            assert next_occ.is_root == was_root

    @pytest.mark.parametrize("include_self", [False, True])
    def test_get_occurrence_start_datetimes(self, occurrences_fixture, include_self, activity_factory):
        single_activity = activity_factory(repeat_choice=Recurrence.ReoccuranceChoice.NEVER, preceded_by=None)
        assert not single_activity._get_occurrence_start_datetimes(include_self)
        assert single_activity.is_root is False

        start_dates = occurrences_fixture[0]._get_occurrence_start_datetimes(include_self)
        nb_occurrences = len(occurrences_fixture) if include_self else len(occurrences_fixture[1:])
        assert len(start_dates) == nb_occurrences

    def test_create_recurrence_child(self, occurrence_fixture):
        nb_activities = Activity.objects.count()
        start_dates = occurrence_fixture._get_occurrence_start_datetimes()
        new_occurrence = occurrence_fixture._create_recurrence_child(start_dates[0])
        assert Activity.objects.count() == nb_activities + 1
        fields = [
            "assigned_to",
            "all_day",
            "conference_room",
            "creator",
            "description",
            "disable_participant_check",
            "importance",
            "visibility",
            "location",
            "location_longitude",
            "location_latitude",
            "recurrence_end",
            "recurrence_count",
            "reminder_choice",
            "repeat_choice",
            "title",
            "type",
        ]
        assert new_occurrence.period.lower == start_dates[0]
        assert occurrence_fixture.duration == new_occurrence.duration
        assert new_occurrence.parent_occurrence == occurrence_fixture
        assert model_to_dict(occurrence_fixture, fields=fields) == model_to_dict(new_occurrence, fields=fields)

    def test_generate_occurrences(self, occurrence_fixture, activity_factory):
        single_activity = activity_factory(repeat_choice=Recurrence.ReoccuranceChoice.NEVER, preceded_by=None)
        assert len(single_activity.generate_occurrences()) == 0
        children = occurrence_fixture.generate_occurrences().order_by("period__startswith")
        assert (
            len(children)
            == len(occurrence_fixture._get_occurrence_start_datetimes())
            == occurrence_fixture.recurrence_count
        )
        assert set(children) == set(occurrence_fixture.child_activities.all())
        assert children[0].is_root is False
        assert len(children[0].generate_occurrences(allow_reclaiming_root=False)) == 0
        assert children[0].is_root is False
        assert len(children[0].generate_occurrences()) == children[0].recurrence_count > 0
        assert children[0].is_root is True
        assert children[0].child_activities.all().count() == children[0].recurrence_count
        assert occurrence_fixture.child_activities.all().count() == 0

    @pytest.mark.parametrize("position_occurrence, propagation", [(0, False), (0, True), (1, False), (1, True)])
    def test_forward_change(self, occurrences_fixture, propagation, position_occurrence, activity_factory):
        activity = occurrences_fixture[position_occurrence]
        nb_next_occurrences = len(occurrences_fixture[position_occurrence + 1 :])
        if activity.is_root:
            assert len(activity.get_recurrent_valid_children()) == nb_next_occurrences
        else:
            assert len(activity.get_recurrent_valid_children()) == 0
        was_root = activity.is_root
        other_activity = activity_factory(preceded_by=None)
        Activity.objects.filter(id=activity.id).update(
            title=other_activity.title, propagate_for_all_children=propagation
        )
        activity.refresh_from_db()
        activity.forward_change()
        activity.refresh_from_db()
        assert activity.is_root == (propagation or was_root)
        assert activity.propagate_for_all_children is False
        if propagation or was_root:
            assert len(activity.get_recurrent_valid_children()) == nb_next_occurrences
        else:
            assert len(activity.get_recurrent_valid_children()) == 0


@pytest.mark.django_db
class TestRecurrenceSerializers(RecurrenceFixture):
    def test_next_occurrence_no_period(self, activity_factory):
        activity = activity_factory(
            period=None, repeat_choice=Recurrence.ReoccuranceChoice.DAILY, recurrence_count=3, preceded_by=None
        )
        request = APIRequestFactory().get("")
        user = UserFactory()
        request.user = user
        request.parser_context = {}
        serializer = ActivityModelSerializer(activity, context={"request": request})
        assert serializer.next_occurrence(activity, request, user)

    def test_previous_occurrence_no_period(self, activity_factory):
        activity = activity_factory(
            period=None, repeat_choice=Recurrence.ReoccuranceChoice.DAILY, recurrence_count=3, preceded_by=None
        )
        request = APIRequestFactory().get("")
        user = UserFactory()
        request.user = user
        request.parser_context = {}
        serializer = ActivityModelSerializer(activity, context={"request": request})
        assert not serializer.previous_occurrence(activity, request, user)

    def test_next_occurrence(self, occurrences_fixture):
        request = APIRequestFactory().get("")
        user = UserFactory()
        request.user = user
        request.parser_context = {}
        last_row = len(occurrences_fixture) - 1
        for row, occurrence in enumerate(occurrences_fixture):
            serializer_occurrence = ActivityModelSerializer(occurrence, context={"request": request})
            if row == last_row:
                assert not serializer_occurrence.next_occurrence(occurrence, request, user)
            else:
                assert serializer_occurrence.next_occurrence(occurrence, request, user) == {
                    "next_occurrence": f"http://testserver/api/crm/activity/{occurrences_fixture[row+1].id}/"
                }

    def test_previous_occurrence(self, occurrences_fixture):
        request = APIRequestFactory().get("")
        user = UserFactory()
        request.user = user
        request.parser_context = {}
        for row, occurrence in enumerate(occurrences_fixture):
            serializer_occurrence = ActivityModelSerializer(occurrence, context={"request": request})
            if row == 0:
                assert not serializer_occurrence.previous_occurrence(occurrence, request, user)
            else:
                assert serializer_occurrence.previous_occurrence(occurrence, request, user) == {
                    "previous_occurrence": f"http://testserver/api/crm/activity/{occurrences_fixture[row-1].id}/"
                }


@pytest.mark.django_db
class TestRecurrenceViewSets(RecurrenceFixture):
    def test_next_and_previous_activity_button_present_for_recurring_activity(
        self, occurrences_fixture, super_user_factory
    ):
        request = APIRequestFactory().get("")
        request.user = super_user_factory()
        request._messages = InMemoryMessageStorage(request)
        view = ActivityViewSet.as_view({"get": "retrieve"})
        for row, occurrence in enumerate(occurrences_fixture):
            response = view(request, pk=occurrence.id).render()
            if occurrence.is_root:
                assert response.data["instance"]["_additional_resources"]["next_occurrence"].endswith(
                    str(occurrences_fixture[row + 1].id) + "/"
                )
                assert "previous_occurrence" not in response.data["instance"]["_additional_resources"]
            elif occurrence.is_leaf:
                assert "next_occurrence" not in response.data["instance"]["_additional_resources"]
                assert response.data["instance"]["_additional_resources"]["previous_occurrence"].endswith(
                    str(occurrences_fixture[row - 1].id) + "/"
                )
            else:
                assert response.data["instance"]["_additional_resources"]["next_occurrence"].endswith(
                    str(occurrences_fixture[row + 1].id) + "/"
                )
                assert response.data["instance"]["_additional_resources"]["previous_occurrence"].endswith(
                    str(occurrences_fixture[row - 1].id) + "/"
                )



from django.utils.translation import gettext as _
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Person
from wbcore.contrib.workflow.decorators import register_assignee
from wbcore.contrib.workflow.models import ProcessStep

# We need to migrate upon name changes!


# TODO: Generalize this through field filter
@register_assignee("Activity Assignee")
def activity_assignee(process_step: ProcessStep, **kwargs) -> User | None:
    if hasattr(process_step.process.instance, "assigned_to"):
        assignee: Person = process_step.process.instance.assigned_to
        if hasattr(assignee, "user_account"):
            return assignee.user_account
        else:
            error_message = _("Assignee has no user account!")
    else:
        error_message = _("No activity attached!")

    process_step.step.get_casted_step().set_failed(
        process_step, _("Error in assignee method: {}").format(error_message)
    )
    return None



from contextlib import suppress
from datetime import date, timedelta
from typing import Iterable

from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateRangeField, RangeOperators
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import Exists, OuterRef, Q, QuerySet, Subquery
from django.db.models.constraints import UniqueConstraint
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from mptt.models import MPTTModel, TreeForeignKey, TreeManager
from psycopg.types.range import DateRange
from slugify import slugify
from wbcore.contrib.ai.llm.decorators import llm
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import EmployerEmployeeRelationship, Entry
from wbcore.contrib.directory.signals import deactivate_profile
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.utils.models import (
    ComplexToStringMixin,
    DeleteToDisableMixin,
)

from wbcrm.models.llm.analyze_relationship import analyze_relationship


class AccountDefaultQueryset(QuerySet):
    def filter_for_user(self, user: User, validity_date: date | None = None, strict: bool = False) -> QuerySet:
        """
        Filters related accounts based on the user's permissions and roles.

        Args:
            user (User): The user for whom related accounts need to be filtered.
            validity_date (date | None, optional): The validity date for role filtering. Defaults to None.
            strict (bool, optional): If True, filtering will be strict based on roles; otherwise, relaxed. Defaults to False.

        Returns:
            QuerySet: A queryset of related accounts filtered based on the user's permissions and roles.
        """
        if user.has_perm("wbcrm.administrate_account"):
            return self
        if not validity_date:
            validity_date = date.today()

        valid_roles = AccountRole.objects.filter_for_user(user, validity_date=validity_date, strict=strict).filter(
            is_currently_valid=True
        )
        if (user.profile.is_internal or user.is_superuser) and not strict:
            return self.filter(Q(id__in=valid_roles.values("account")) | Q(is_public=True))
        return self.annotate(
            has_direct_role=Exists(valid_roles.filter(account=OuterRef("id"))),
            has_descending_roles=Exists(
                valid_roles.filter(
                    account__tree_id=OuterRef("tree_id"),
                    account__lft__lte=OuterRef("lft"),
                    account__rght__gte=OuterRef("rght"),
                )
            ),
        ).filter((Q(is_public=True) & Q(has_descending_roles=True)) | Q(has_direct_role=True))


class AccountManager(models.Manager):
    def get_queryset(self) -> AccountDefaultQueryset:
        return AccountDefaultQueryset(self.model)

    def filter_for_user(self, user: User, validity_date: date | None = None, strict: bool = False) -> QuerySet:
        return self.get_queryset().filter_for_user(user, validity_date=validity_date, strict=strict)


class ActiveAccountManager(AccountManager):
    def get_queryset(self):
        return super().get_queryset().filter(is_active=True)


class OpenAccountObjectManager(AccountManager):
    def get_queryset(self):
        return super().get_queryset().filter(status=Account.Status.OPEN, is_active=True)


@llm([analyze_relationship])
class Account(ComplexToStringMixin, DeleteToDisableMixin, WBModel, MPTTModel):
    tree_id: int

    class Status(models.TextChoices):
        PENDING = "PENDING", _("Pending")
        OPEN = "OPEN", _("Open")
        CLOSE = "CLOSE", _("Close")

    relationship_status = models.PositiveIntegerField(
        validators=[MinValueValidator(1), MaxValueValidator(5)],
        verbose_name=_("Relationship Status"),
        help_text=_("The Relationship Status from 1 to 5. 1 being the cold and 5 being the hot."),
        blank=True,
        null=True,
    )
    relationship_summary = models.TextField(default="", blank=True)
    action_plan = models.TextField(default="", blank=True)

    reference_id = models.PositiveIntegerField(unique=True, blank=True)
    title = models.CharField(max_length=255, verbose_name="Title")
    status = FSMField(default=Status.OPEN, choices=Status.choices, verbose_name="Status")
    parent = TreeForeignKey(
        "wbcrm.Account",
        related_name="children",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        verbose_name=_("Parent Account"),
    )
    is_terminal_account = models.BooleanField(
        default=False,
        verbose_name="Terminal Account",
        help_text="If true, sales or revenue can happen in this account",
    )
    is_public = models.BooleanField(
        default=True, verbose_name="Public", help_text="If True, all internal users can access this account"
    )
    owner = models.ForeignKey(
        "directory.Entry",
        related_name="accounts",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("Owner"),
    )

    @transition(
        status,
        Status.PENDING,
        Status.OPEN,
        permission=lambda account, user: account.can_administrate(user),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:account",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                description_fields="<p>Are you sure you want to open this account?</p>",
            )
        },
    )
    def approve(self, **kwargs):
        pass

    @transition(
        status,
        Status.PENDING,
        Status.CLOSE,
        permission=lambda account, user: account.can_administrate(user),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:account",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                description_fields="<p>Are you sure you want to close this account?</p>",
            )
        },
    )
    def deny(self, **kwargs):
        pass

    @transition(
        status,
        Status.OPEN,
        Status.CLOSE,
        permission=lambda account, user: account.can_administrate(user),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:account",),
                icon=WBIcon.LOCK.icon,
                key="close",
                label="Close",
                action_label="Close",
                description_fields="<p>Are you sure you want to close this account?</p>",
            )
        },
    )
    def close(self, **kwargs):
        pass

    @transition(
        status,
        Status.CLOSE,
        Status.PENDING,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcrm:account",),
                icon=WBIcon.LOCK.icon,
                key="reopen",
                label="Reopen",
                action_label="Reopen",
                description_fields="<p>Are you sure you want to reopen this account?</p>",
            )
        },
    )
    def reopen(self, **kwargs):
        pass

    def can_administrate(self, user: User):
        """Every superuser, valid manager and valid pm can lock a"""
        return user.has_perm("wbcrm.administrate_account")

    def compute_str(self) -> str:
        title = self.title or str(self.reference_id)
        if self.parent:
            title += f" ({self.parent.computed_str})"
        return title

    def merge(self, merged_account):
        if not merged_account.children.exists() and (
            merged_account.parent == self or merged_account.parent == self.parent
        ):  # we can merge only sibling accounts or child to parent
            with transaction.atomic():  # We want this to either succeed fully or fail
                for role in merged_account.roles.all():
                    try:
                        new_role = AccountRole.objects.get(entry=role.entry, account=self)
                        for validity in role.validity_set.all():
                            if not AccountRoleValidity.objects.filter(
                                role=new_role, timespan__overlap=validity.timespan
                            ).exists():
                                validity.role = new_role
                                validity.save()
                        for user in role.authorized_hidden_users.all():
                            new_role.authorized_hidden_users.add(user)
                        role.delete()
                    except AccountRole.DoesNotExist:
                        role.account = self
                        role.save()

                # Get the base
                pre_merge.send(
                    sender=Account, merged_object=merged_account, main_object=self
                )  # default signal dispatch for the Account class
                # We delete finally the merged account. All unlikage should have been done in the signal receivers function ( we refresh to be sure that no receiver modified the given object )
                self.refresh_from_db()
                merged_account.refresh_from_db()
                merged_account.delete(no_deletion=False)

                # copy fields

            # trigger save for post save logic (if any)
            self.save()

    def save(self, *args, **kwargs):
        if self.parent and not self.owner:
            self.owner = self.parent.owner
        if not self.reference_id:
            self.reference_id = Account.get_next_available_reference_id()
        # if not Account.objects.filter(parent=self).exists():
        #     self.is_terminal_account = True
        # else:
        #     self.is_terminal_account = False
        self.is_terminal_account = self.is_leaf_node()
        if not self.is_active:
            self.status = self.Status.CLOSE
        if self.status == self.Status.CLOSE:
            self.is_active = False
        super().save(*args, **kwargs)
        # self.children.update() TODO recompute str for all children
        # Account.objects.filter(id=self.id).update(computed_str=self.compute_str())

    def get_inherited_roles_for_account(self, include_self: bool = False) -> QuerySet:
        """
        Return account role from the parent accounts

        Args:
            include_self: MPTT argument

        Returns:
            The parent account roles
        """
        return AccountRole.objects.filter(account__in=self.get_ancestors(include_self=include_self))

    def can_see_account(self, user: User, validity_date: date | None = None) -> bool:
        """
        Checks if the user can see the account based on their permissions and roles.

        Args:
            user (User): The user for whom account visibility needs to be checked.
            validity_date (date | None, optional): The validity date for role filtering. Defaults to None.

        Returns:
            bool: True if the user can see the account, False otherwise.
        """
        if not validity_date:
            validity_date = date.today()
        return Account.objects.filter(id=self.id).filter_for_user(user, validity_date=validity_date).exists()

    @classmethod
    def get_next_available_reference_id(cls) -> int:
        if Account.objects.exists():
            reference_id = Account.all_objects.latest("reference_id").reference_id + 1
        else:
            reference_id = 1
        return reference_id

    @classmethod
    def annotate_root_account_info(cls, queryset: QuerySet) -> QuerySet:
        """
        Utility classmethod to annotate a queryset for the root account and its owner

        Args:
            queryset: Queryset to annotate

        Returns:
            A annotated queryset
        """
        return queryset.annotate(
            root_account=Subquery(
                Account.all_objects.filter(tree_id=OuterRef("account__tree_id"), level=0).values("id")[:1]
            ),
            root_account_repr=Subquery(
                Account.all_objects.filter(tree_id=OuterRef("account__tree_id"), level=0).values("computed_str")[:1]
            ),
            root_account_owner=Subquery(
                Account.all_objects.filter(tree_id=OuterRef("account__tree_id"), level=0).values("owner")[:1]
            ),
            root_account_owner_repr=Subquery(
                Account.all_objects.filter(tree_id=OuterRef("account__tree_id"), level=0).values(
                    "owner__computed_str"
                )[:1]
            ),
        )

    class Meta:
        verbose_name = _("Account")
        verbose_name_plural = _("Accounts")
        permissions = [("administrate_account", "Administrate Account")]

    objects = ActiveAccountManager()
    all_objects = AccountManager()
    open_objects = OpenAccountObjectManager()
    tree_objects = TreeManager()

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcrm:account"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcrm:accountrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_accounts_for_customer(cls, entries: Entry | Iterable[Entry]) -> QuerySet:
        """
        Retrieves accounts associated with the given entry's ownership.

        Args:
            entries (Entry): The entry for which owned accounts are to be retrieved. Can be an iterable or an entry.

        Returns:
            QuerySet: A queryset of accounts owned by the provided entry.
        """
        if not isinstance(entries, Iterable):
            entries = [entries]

        # Get all root accounts owned by the entry
        root_accounts = cls.objects.filter(
            Q(owner__in=entries)
            | Q(
                owner__in=EmployerEmployeeRelationship.objects.filter(employee__id__in=[o.id for o in entries]).values(
                    "employer"
                )
            )
        )

        # Get root account and descendants account ids
        return cls.objects.annotate(
            is_direct_owner=Exists(root_accounts.filter(id=OuterRef("id"))),
            is_owner_of_descending_account=Exists(
                root_accounts.filter(tree_id=OuterRef("tree_id"), lft__lte=OuterRef("lft"), rght__gte=OuterRef("rght"))
            ),
        ).filter(Q(is_owner_of_descending_account=True) | Q(is_direct_owner=True))

    @classmethod
    def get_managed_accounts_for_entry(cls, entry: Entry) -> QuerySet:
        """
        Retrieves managed accounts associated with the given entry.

        Args:
            entry (Entry): The entry for which managed accounts are to be retrieved.

        Returns:
            QuerySet: A queryset of managed accounts associated with the provided entry.
        """
        roles = AccountRole.objects.filter(entry=entry)

        return cls.objects.annotate(
            has_direct_role=Exists(roles.filter(account=OuterRef("id"))),
            has_descending_roles=Exists(
                roles.filter(
                    account__tree_id=OuterRef("tree_id"),
                    account__lft__lte=OuterRef("lft"),
                    account__rght__gte=OuterRef("rght"),
                )
            ),
        ).filter((Q(is_public=True) & Q(has_descending_roles=True)) | Q(has_direct_role=True))


class AccountRoleType(models.Model):
    title = models.CharField(max_length=126, verbose_name="Title")
    key = models.CharField(max_length=126, unique=True)

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = slugify(self.title)
        super().save(*args, **kwargs)

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcrm:accountroletyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


class AccountRoleDefaultQueryset(QuerySet):
    def filter_for_user(self, user: User, validity_date: date | None = None, strict: bool = False) -> QuerySet:
        """
        Filters account roles related to the user based on permissions and roles.

        Args:
            user (User): The user for whom account roles need to be filtered.
            validity_date (date | None, optional): The validity date for role filtering. Defaults to None.
            strict (bool, optional): If True, filtering will be strict based on roles; otherwise, relaxed. Defaults to False.

        Returns:
            QuerySet: A queryset of account roles related to the user, filtered based on permissions and roles.
        """
        if not validity_date:
            validity_date = date.today()
        qs = self.annotate(is_currently_valid=AccountRoleValidity.get_role_validity_subquery(validity_date))
        if user.has_perm("wbcrm.administrate_account"):
            return qs
        if user.profile.is_internal and not strict:
            qs = qs.filter(
                Q(entry_id=user.profile.id)
                | (Q(account__is_public=True) & Q(is_hidden=False))
                | (Q(is_hidden=True) & Q(authorized_hidden_users=user))
            )
        else:
            qs = qs.filter(entry_id=user.profile.id)
        return qs


class AccountRoleManager(models.Manager):
    # Necessary because otherwise pyright cannot find method

    def get_queryset(self) -> AccountRoleDefaultQueryset:
        return AccountRoleDefaultQueryset(self.model)

    def filter_for_user(self, user: User, validity_date: date | None = None, strict: bool = False) -> QuerySet:
        return self.get_queryset().filter_for_user(user, validity_date=validity_date, strict=strict)


class AccountRole(ComplexToStringMixin):
    """Model for Account Roles"""

    class Meta:
        verbose_name = "Account Role"
        verbose_name_plural = "Account Roles"
        constraints = [UniqueConstraint(fields=["account", "entry"], name="unique_account_entry_relationship")]

    role_type = models.ForeignKey(
        "wbcrm.AccountRoleType", related_name="roles", on_delete=models.PROTECT, verbose_name="Role Type"
    )
    entry = models.ForeignKey(
        "directory.Entry", related_name="account_roles", on_delete=models.PROTECT, verbose_name="Entry"
    )
    account = models.ForeignKey(
        "wbcrm.Account", related_name="roles", on_delete=models.CASCADE, verbose_name="Account"
    )

    is_hidden = models.BooleanField(
        default=False,
        verbose_name="Hidden",
        help_text="If True, this role is hidden and can be seen only by authorized people",
    )
    authorized_hidden_users = models.ManyToManyField(
        "authentication.User",
        related_name="authorized_hidden_roles",
        blank=True,
        verbose_name=_("authorized Hidden Users"),
        help_text=_("List of users that are allowed to see this hidden account role"),
    )

    weighting = models.FloatField(default=1, verbose_name="Weight")

    objects = AccountRoleManager()

    def compute_str(self) -> str:
        rel = f"Role {self.role_type} for {self.entry} on {self.account}"
        if self.is_hidden:
            rel += " (Hidden)"
        return rel

    def save(self, *args, **kwargs):
        # if the role is hidden and the account is public, we ensure it becomes private so that the hidden rule is respected
        if self.is_hidden and self.account.is_public:
            self.account.is_public = False
            self.account.save()

        super().save(*args, **kwargs)

    def deactivate(self, deactivation_date: date | None = None):
        """
        Utility function to disable a account role at a given time

        Args:
            deactivation_date: The time at which the role will be deactivated. Default to today
        """
        if not deactivation_date:
            deactivation_date = date.today()
        with suppress(AccountRoleValidity.DoesNotExist):
            val = AccountRoleValidity.objects.get(
                role=self,
                timespan__startswith__lte=deactivation_date,
                timespan__endswith__gt=deactivation_date,
            )
            val.timespan = DateRange(val.timespan.lower, deactivation_date)
            val.save()


class AccountRoleValidity(models.Model):
    role = models.ForeignKey(
        "wbcrm.AccountRole", related_name="validity_set", on_delete=models.CASCADE, verbose_name="Account Role"
    )
    timespan = DateRangeField(verbose_name="Timespan")

    class Meta:
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_roles",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("role", RangeOperators.EQUAL),
                ],
            ),
        ]

    def __str__(self):
        return f"[{self.timespan.lower} - {self.timespan.upper}["  # type: ignore

    @classmethod
    def get_role_validity_subquery(cls, validity_date: date, role_label_key: str = "pk") -> Subquery:
        """
        Return a subquery that will define wether a account role is valid
        Args:
            validity_date: The validity date
            role_label_key: The related name for the account role foreign key

        Returns:
            A subquery expression of type boolean
        """
        return Exists(
            AccountRoleValidity.objects.filter(
                role=OuterRef(role_label_key),
                timespan__startswith__lte=validity_date,
                timespan__endswith__gt=validity_date,
            )
        )


@receiver(deactivate_profile)
def handle_user_deactivation(sender, instance, substitute_profile=None, **kwargs):
    deactivation_date = date.today() - timedelta(days=1)
    for profile_role in AccountRole.objects.filter(entry_id=instance.id):
        for validity in profile_role.validity_set.all():
            if validity.timespan.upper >= deactivation_date:  # type: ignore
                validity.timespan = DateRange(
                    validity.timespan.lower,
                    max([deactivation_date, validity.timespan.lower]),  # type: ignore
                )
                validity.save()
                if substitute_profile and validity.timespan.lower <= deactivation_date:  # type: ignore
                    substitute_role, created = AccountRole.objects.get_or_create(
                        account=profile_role.account,
                        entry_id=substitute_profile.id,
                        defaults={"role_type": profile_role.role_type},
                    )
                    if created:
                        v = substitute_role.validity_set.filter(
                            timespan__startswith__lt=deactivation_date,
                            timespan__endswith__gt=deactivation_date,
                        ).first()
                        v.timespan = DateRange(deactivation_date, date.max)  # type: ignore
                        v.save()


@receiver(post_save, sender="wbcrm.Account")
def post_account_creation(sender, instance, created, **kwargs):
    # disabling parent account disable children as well
    if not instance.is_active:
        instance.get_descendants().update(is_active=False)
    # check that if an account is private, all its children are private as well
    if not instance.is_public:
        instance.get_descendants().update(is_public=False)
    # if an new account is created and it's a leaf node, we assume it's a terminal account. Can be changed afterwards
    if created:
        # we create a role for the owner by default upon creation
        if instance.owner:
            owner_role_type = AccountRoleType.objects.get_or_create(key="customer", defaults={"title": "Customer"})[0]
            AccountRole.objects.get_or_create(
                account=instance, entry=instance.owner, defaults={"role_type": owner_role_type}
            )
        if instance.is_terminal_account and instance.parent:
            instance.get_ancestors().update(is_terminal_account=False)


@receiver(post_delete, sender="wbcrm.Account")
def post_delete_account(sender, instance, **kwargs):
    if (parent := instance.parent) and not parent.children.exists():
        parent.is_terminal_account = True
        parent.save()


@receiver(post_save, sender="wbcrm.AccountRole")
def post_account_role_creation(sender, instance, created, **kwargs):
    # if an new account is created and it's a leaf node, we assume it's a terminal account. Can be changed afterwards
    if created:
        AccountRoleValidity.objects.create(role=instance, timespan=DateRange(date.min, date.max))  # type: ignore



from django.dispatch import receiver
from django.utils.translation import gettext as _
from wbcore.contrib.directory.viewsets import (
    CompanyModelViewSet,
    EntryModelViewSet,
    PersonModelViewSet,
)
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.signals.instance_buttons import add_instance_button


@receiver(add_instance_button, sender=PersonModelViewSet)
@receiver(add_instance_button, sender=EntryModelViewSet)
@receiver(add_instance_button, sender=CompanyModelViewSet)
def add_activity_instance_button_in_directory_viewsets(sender, many, *args, **kwargs):
    return bt.WidgetButton(key="activity", label=_("Activities"), icon=WBIcon.CALENDAR.icon, weight=1)



from .accounts import Account, AccountRole, AccountRoleType, AccountRoleValidity
from .activities import Activity, ActivityParticipant, ActivityType, add_employer_to_activities
from .groups import Group
from .products import Product, ProductCompanyRelationship
from .events import Event



# import datetime
# from zoneinfo import ZoneInfo

# import pytest
# from django.conf import settings
# from dynamic_preferences.registries import global_preferences_registry
# from wbcrm.models import (
#     Activity,
#     ActivityParticipant,
#     CalendarItem,
#     EmailContact,
#     Person,
# )

# from ..utils import GoogleSyncUtils
# from .test_data import event, event_list


# @pytest.mark.django_db
# class TestUtils:
#     def test_add_instance_metadata(self, activity_factory):
#         new_metadata = {"google_backend": {"instance": {"uID": "UID"}, "event": {"id": "event id"}}}
#         parent_occurrence: Activity = activity_factory(
#             metadata={"google_backend": {"instance": {"originalStartTime": {"dateTime": "Fake Date Time"}}}}
#         )
#         child_activity_a: Activity = activity_factory(
#             parent_occurrence=parent_occurrence,
#             metadata={"google_backend": {"instance": {"originalStartTime": {"dateTime": "Fake Date Time A"}}}},
#         )
#         child_activity_b: Activity = activity_factory(
#             parent_occurrence=parent_occurrence,
#             metadata={"google_backend": {"instance": {"originalStartTime": {"dateTime": "Fake Date Time B"}}}},
#         )
#         child_activity_c: Activity = activity_factory(
#             parent_occurrence=parent_occurrence,
#             metadata={"google_backend": {"instance": {"originalStartTime": {"dateTime": "Fake Date Time C"}}}},
#         )
#         GoogleSyncUtils.add_instance_metadata(parent_occurrence, event_list, new_metadata)
#         parent_occurrence = Activity.objects.get(id=parent_occurrence.id)
#         child_activity_a = Activity.objects.get(id=child_activity_a.id)
#         child_activity_b = Activity.objects.get(id=child_activity_b.id)
#         child_activity_c = Activity.objects.get(id=child_activity_c.id)
#         assert parent_occurrence.metadata["google_backend"]["instance"]["metaTest"] == "Parent"
#         assert child_activity_a.external_id == "2"
#         assert child_activity_a.metadata["google_backend"]["instance"]["metaTest"] == "Child A"
#         assert child_activity_b.external_id == "3"
#         assert child_activity_b.metadata["google_backend"]["instance"]["metaTest"] == "Child B"
#         assert child_activity_c.external_id == "4"
#         assert child_activity_c.metadata["google_backend"]["instance"]["metaTest"] == "Child C"

#     def test_convert_activity_visibility_to_event_visibility(self):
#         assert (
#             GoogleSyncUtils.convert_activity_visibility_to_event_visibility(CalendarItem.Visibility.PUBLIC) == "public"
#         )
#         assert (
#             GoogleSyncUtils.convert_activity_visibility_to_event_visibility(CalendarItem.Visibility.PRIVATE)
#             == "private"
#         )
#         assert (
#             GoogleSyncUtils.convert_activity_visibility_to_event_visibility(CalendarItem.Visibility.CONFIDENTIAL)
#             == "private"
#         )

#     def test_convert_event_visibility_to_activity_visibility(self):
#         assert (
#             GoogleSyncUtils.convert_event_visibility_to_activity_visibility("private")
#             == CalendarItem.Visibility.PRIVATE
#         ) or (
#             GoogleSyncUtils.convert_event_visibility_to_activity_visibility("private")
#             == CalendarItem.Visibility.CONFIDENTIAL
#         )
#         assert (
#             GoogleSyncUtils.convert_event_visibility_to_activity_visibility("public") == CalendarItem.Visibility.PUBLIC
#         )

#     def test_convert_attendee_status_to_participant_status(self):
#         assert (
#             GoogleSyncUtils.convert_attendee_status_to_participant_status("accepted")
#             == ActivityParticipant.ParticipationStatus.ATTENDS
#         )
#         assert (
#             GoogleSyncUtils.convert_attendee_status_to_participant_status("declined")
#             == ActivityParticipant.ParticipationStatus.CANCELLED
#         )
#         assert (
#             GoogleSyncUtils.convert_attendee_status_to_participant_status("tentative")
#             == ActivityParticipant.ParticipationStatus.MAYBE
#         )
#         assert (
#             GoogleSyncUtils.convert_attendee_status_to_participant_status("Something different")
#             == ActivityParticipant.ParticipationStatus.NOTRESPONDED
#         )

#     def test_convert_participant_status_to_attendee_status(self):
#         assert (
#             GoogleSyncUtils.convert_participant_status_to_attendee_status(
#                 ActivityParticipant.ParticipationStatus.ATTENDS
#             )
#             == "accepted"
#         )
#         assert (
#             GoogleSyncUtils.convert_participant_status_to_attendee_status(
#                 ActivityParticipant.ParticipationStatus.ATTENDS_DIGITALLY
#             )
#             == "accepted"
#         )
#         assert (
#             GoogleSyncUtils.convert_participant_status_to_attendee_status(
#                 ActivityParticipant.ParticipationStatus.CANCELLED
#             )
#             == "declined"
#         )
#         assert (
#             GoogleSyncUtils.convert_participant_status_to_attendee_status(
#                 ActivityParticipant.ParticipationStatus.MAYBE
#             )
#             == "tentative"
#         )
#         assert GoogleSyncUtils.convert_participant_status_to_attendee_status("Something different") == "notResponded"

#     @pytest.mark.parametrize("google_event", [{}, event])
#     def test_get_participants(self, google_event, person_factory, email_contact_factory):
#         person_a = person_factory()
#         person_b = person_factory()
#         person_c = person_factory()
#         email_contact_factory(primary=True, entry=person_a, address="Foo@Foo.com")
#         email_contact_factory(primary=True, entry=person_b, address="Bar@Bar.com")
#         part_list = GoogleSyncUtils.get_or_create_participants(google_event, person_c)
#         if not google_event:
#             assert part_list == [{"person_id": person_c.id, "status": ActivityParticipant.ParticipationStatus.ATTENDS}]
#         else:
#             person_d = Person.objects.get(last_name="Foo Bar")
#             person_e = Person.objects.get(last_name="Bar@Foo.com")
#             assert len(part_list) == 5
#             assert {"person_id": person_a.id, "status": ActivityParticipant.ParticipationStatus.ATTENDS} in part_list
#             assert {"person_id": person_b.id, "status": ActivityParticipant.ParticipationStatus.CANCELLED} in part_list
#             assert {"person_id": person_c.id, "status": ActivityParticipant.ParticipationStatus.ATTENDS} in part_list
#             assert {"person_id": person_d.id, "status": ActivityParticipant.ParticipationStatus.MAYBE} in part_list
#             assert {"person_id": person_e.id, "status": ActivityParticipant.ParticipationStatus.MAYBE} in part_list

#     @pytest.mark.parametrize("can_sync", [False, True])
#     @pytest.mark.parametrize("google_event", [{}, event])
#     def test_convert_activity_participants_to_attendees_list(
#         self, can_sync, google_event, activity_factory, person_factory, internal_user_factory, email_contact_factory
#     ):
#         global_preferences_registry.manager()["wbactivity_sync__sync_external_participants"] = can_sync
#         person_a: Person = internal_user_factory()
#         person_b: Person = person_factory()
#         person_c: Person = person_factory()
#         email_contact_factory(primary=True, entry=person_a, address="A@A.com")
#         email_contact_factory(primary=True, entry=person_b, address="B@B.com")
#         email_contact_factory(primary=True, entry=person_c, address="Foo@Bar.com")
#         activity: Activity = activity_factory(creator=person_a, participants=[person_a.id, person_b.id, person_c])
#         # attendees: list[dict] = convert_activity_participants_to_attendees_list(activity, google_event)
#         attendees: list[dict] = GoogleSyncUtils.convert_activity_participants_to_attendees(activity, google_event)

#         person_a_dict: dict = {
#             "displayName": person_a.computed_str,
#             "email": str(person_a.primary_email_contact()),
#             "responseStatus": "accepted",
#         }
#         person_b_dict: dict = {
#             "displayName": person_b.computed_str,
#             "email": str(person_b.primary_email_contact()),
#             "responseStatus": "needsAction",
#         }
#         person_c_dict: dict = {
#             "displayName": person_c.computed_str,
#             "email": str(person_c.primary_email_contact()),
#             "responseStatus": "needsAction",
#         }
#         if not google_event and not can_sync:
#             assert len(attendees) == 1
#             assert person_a_dict in attendees
#         elif not google_event and can_sync:
#             assert len(attendees) == 3
#             assert person_a_dict in attendees
#             assert person_b_dict in attendees
#         elif google_event and not can_sync:
#             assert len(attendees) == 2
#             assert person_a_dict in attendees
#             assert person_c_dict in attendees
#             assert person_b_dict not in attendees
#         else:
#             assert len(attendees) == 3
#             assert person_a_dict in attendees
#             assert person_b_dict in attendees
#             assert person_c_dict in attendees

#     @pytest.mark.parametrize("created", [False])
#     @pytest.mark.parametrize(
#         "metadata",
#         [{}, {"recurrence": ["Recurrence A"], "google_backend": {"event": {"recurrence": ["Recurrence B"]}}}],
#     )
#     def test_convert_activity_to_event(
#         self, internal_user_factory, activity_factory, email_contact_factory, created, metadata
#     ):
#         timezone = ZoneInfo(settings.TIME_ZONE)
#         person_a: Person = internal_user_factory()
#         person_b: Person = internal_user_factory()
#         email_contact_factory(primary=True, entry=person_a, address="Foo@Foo.com")
#         email_contact_factory(primary=True, entry=person_b, address="Bar@Bar.com")
#         activity: Activity = activity_factory(
#             creator=person_a,
#             participants=[person_b.id],
#             visibility=CalendarItem.Visibility.PUBLIC,
#             metadata=metadata,
#         )
#         event = GoogleSyncUtils.convert_activity_to_event(activity, created)
#         assert event["summary"] == activity.title
#         assert event["creator"] == str(activity.creator.primary_email_contact())
#         assert event["organizer"] == str(activity.assigned_to.primary_email_contact())
#         assert event["location"] == activity.location
#         assert event["visibility"] == "public"
#         assert event["attendees"] == [
#             {
#                 "displayName": person_b.computed_str,
#                 "email": str(person_b.primary_email_contact()),
#                 "responseStatus": "needsAction",
#             }
#         ]
#         assert event["start"] == {
#             "dateTime": activity.period.lower.astimezone(timezone).isoformat(),
#             "timeZone": settings.TIME_ZONE,
#         }
#         assert event["end"] == {
#             "dateTime": activity.period.upper.astimezone(timezone).isoformat(),
#             "timeZone": settings.TIME_ZONE,
#         }
#         if metadata and not created:
#             assert event["recurrence"] == ["Recurrence B"]
#         elif metadata and created:
#             assert event["recurrence"] == ["Recurrence A"]
#         else:
#             assert event["recurrence"] == []

#     def test_get_start_and_end(self):
#         meta = {
#             "start": {"dateTime": "2022-12-20T11:30:00+01:00", "timeZone": "Europe/Berlin"},
#             "end": {"dateTime": "2022-12-20T12:30:00+01:00", "timeZone": "Europe/Berlin"},
#         }
#         meta_date = {
#             "start": {"date": "2022-12-24"},
#             "end": {"date": "2022-12-25"},
#         }
#         start, end = GoogleSyncUtils.get_start_and_end(meta)
#         assert start == datetime.datetime(
#             2022, 12, 20, 11, 30, tzinfo=datetime.timezone(datetime.timedelta(seconds=3600))
#         )
#         assert end == datetime.datetime(
#             2022, 12, 20, 12, 30, tzinfo=datetime.timezone(datetime.timedelta(seconds=3600))
#         )
#         start, end = GoogleSyncUtils.get_start_and_end(meta_date)
#         assert start == datetime.datetime(2022, 12, 24)
#         assert end == datetime.datetime(2022, 12, 24, 23, 59, 59)

#     @pytest.mark.parametrize(
#         "organizer",
#         ["", "Foo@Foo.com", "Bar@Bar.com"],
#     )
#     def test_get_person(self, person_factory, email_contact_factory, organizer):
#         person_a: Person = person_factory()
#         email_contact_factory(primary=True, entry=person_a, address="foo@foo.com")

#         if organizer == "Foo@Foo.com":
#             assert GoogleSyncUtils.get_or_create_person(organizer) == person_a
#         elif organizer and (mail := organizer) == "Bar@Bar.com":
#             creator = GoogleSyncUtils.get_or_create_person(mail)
#             person_b = Person.objects.get(last_name=mail)
#             assert EmailContact.objects.filter(entry=creator, address=mail.lower(), primary=True).exists()
#             assert creator.id == person_b.id
#             assert creator.last_name == mail
#         else:
#             assert GoogleSyncUtils.get_or_create_person(organizer) is None



import factory

from wbcrm.models import Product


class ProductFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("text", max_nb_chars=32)
    is_competitor = factory.Faker("pybool")

    class Meta:
        model = Product



from datetime import date, timedelta

import django_filters
from django.contrib.auth import get_user_model
from django.db.models import Exists, OuterRef, Q
from django.db.models.query import QuerySet
from django.utils import timezone
from django.utils.translation import gettext_lazy
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from psycopg.types.range import DateRange, TimestamptzRange
from wbcore import filters as wb_filters
from wbcore.contrib.agenda.filters import CalendarItemPeriodBaseFilterSet
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.directory.models import CustomerStatus
from wbcore.contrib.directory.models.entries import Company, Person
from wbcore.filters.lookups import ALL_TEXT_LOOKUPS

from wbcrm.models import Activity, ActivityParticipant, ActivityType, Group
from wbcrm.models.activities import ActivityCompanyThroughModel


def get_employee_filter_params(request, view):
    if employer_id := global_preferences_registry.manager()["directory__main_company"]:
        return {"employers": employer_id}
    return {}


class ActivityBaseFilterSet(CalendarItemPeriodBaseFilterSet):
    type = wb_filters.ModelMultipleChoiceFilter(
        label=gettext_lazy("Types"),
        queryset=ActivityType.objects.all(),
        endpoint=ActivityType.get_representation_endpoint(),
        value_key=ActivityType.get_representation_value_key(),
        label_key=ActivityType.get_representation_label_key(),
    )

    visibility = wb_filters.MultipleChoiceFilter(
        choices=CalendarItem.Visibility.choices,
        label=gettext_lazy("Visibility"),
    )

    participants = wb_filters.ModelMultipleChoiceFilter(
        label=gettext_lazy("Participants"),
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
    )

    companies = wb_filters.ModelMultipleChoiceFilter(
        label=gettext_lazy("Companies"),
        queryset=Company.objects.all(),
        endpoint=Company.get_representation_endpoint(),
        value_key=Company.get_representation_value_key(),
        label_key=Company.get_representation_label_key(),
    )

    groups = wb_filters.ModelMultipleChoiceFilter(
        label=gettext_lazy("Groups"),
        queryset=Group.objects.all(),
        endpoint=Group.get_representation_endpoint(),
        value_key=Group.get_representation_value_key(),
        label_key=Group.get_representation_label_key(),
    )


class ActivityFilter(ActivityBaseFilterSet):
    clients_of = wb_filters.ModelChoiceFilter(
        label=gettext_lazy("Clients of"),
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
        method="filter_clients_of",
        filter_params=get_employee_filter_params,
    )

    latest_reviewer = wb_filters.ModelMultipleChoiceFilter(
        label=gettext_lazy("Latest Reviewers"),
        queryset=Person.objects.all(),
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
    )

    importance = wb_filters.MultipleChoiceFilter(
        label=gettext_lazy("Importance"), choices=Activity.Importance.choices, widget=django_filters.widgets.CSVWidget
    )

    created = wb_filters.DateTimeRangeFilter(label=gettext_lazy("Created"))
    edited = wb_filters.DateTimeRangeFilter(label=gettext_lazy("Edited"))

    only_recent = wb_filters.BooleanFilter(label=gettext_lazy("Only Recent"), method="boolean_only_recent")

    customer_status = wb_filters.ModelChoiceFilter(
        label=gettext_lazy("Customer Status"),
        queryset=CustomerStatus.objects.all(),
        endpoint=CustomerStatus.get_representation_endpoint(),
        value_key=CustomerStatus.get_representation_value_key(),
        label_key=CustomerStatus.get_representation_label_key(),
        method="filter_customer_status",
        help_text="Filter activities based on the company customer status at the activity creation time",
    )

    def boolean_only_recent(self, queryset: QuerySet[Activity], name, value: bool | None) -> QuerySet[Activity]:
        if value:
            today = date.today()
            next_week = today + timedelta(days=7)
            last_month = today - timedelta(days=30)
            return queryset.filter(period__overlap=TimestamptzRange(last_month, next_week))
        return queryset

    def filter_clients_of(self, queryset, name, value):
        if value:
            user = self.request.user
            return Activity.objects.filter(
                (Q(participants__relationship_managers=value) | Q(companies__relationship_managers=value))
                & (
                    Q(visibility=CalendarItem.Visibility.PUBLIC)
                    | (
                        Q(visibility=CalendarItem.Visibility.CONFIDENTIAL)  # TODO move that to a queryset method
                        & Exists(
                            get_user_model().objects.filter(
                                Q(id=user.id)
                                & (
                                    Q(groups__permissions__codename="administrate_confidential_items")
                                    | Q(user_permissions__codename="administrate_confidential_items")
                                )
                            )
                        )
                    )
                    | Q(
                        Exists(
                            ActivityParticipant.objects.filter(activity_id=OuterRef("pk"), participant=user.profile)
                        )
                    )
                    | Q(assigned_to=user.profile)
                )
            )
        return queryset

    def filter_customer_status(self, queryset, name, value):
        if value:
            rel = ActivityCompanyThroughModel.objects.filter(customer_status=value)
            return queryset.filter(id__in=rel.values("activity"))
        return queryset

    class Meta:
        model = Activity
        fields = {
            "status": ["exact"],
            "repeat_choice": ["exact"],
            "title": ALL_TEXT_LOOKUPS,
            "result": ALL_TEXT_LOOKUPS,
            "description": ALL_TEXT_LOOKUPS,
        }


def default_activitychart_get_params(*args, **kwargs):
    current_last_week_date_start = timezone.now().date() - timedelta(days=7)
    current_next_week_date_end = timezone.now().date() + timedelta(days=7)
    return DateRange(current_last_week_date_start, current_next_week_date_end)


class ActivityChartFilter(ActivityBaseFilterSet):
    status = wb_filters.MultipleChoiceFilter(
        label=gettext_lazy("Status"), choices=Activity.Status.choices, widget=django_filters.widgets.CSVWidget
    )
    period = wb_filters.DateTimeRangeFilter(label="Period", required=True, initial=default_activitychart_get_params)

    class Meta:
        model = Activity
        fields = {}


class ActivityTypeFilter(wb_filters.FilterSet):
    score = wb_filters.MultipleChoiceFilter(
        label=_("Multipliers"), choices=ActivityType.Score.choices[:3], widget=django_filters.widgets.CSVWidget
    )

    class Meta:
        model = ActivityType
        fields = {
            "title": ["exact", "icontains"],
            "color": ["exact", "icontains"],
            "default": ["exact"],
        }


class ActivityParticipantFilter(wb_filters.FilterSet):
    is_occupied_filter = wb_filters.BooleanFilter(
        label=gettext_lazy("Is occupied by different activity"), method="boolean_is_occupied"
    )

    def boolean_is_occupied(self, queryset, name, value):
        if value is None:
            return queryset
        else:
            return queryset.filter(is_occupied=value)

    class Meta:
        model = ActivityParticipant
        fields = {}
